/**************************************************************************************

                                    nDB Mapper
                           Database Memory Manipulation

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "mapper.h"

#include <stdlib.h>
#include <string.h>

//-------------------------------------------------------------------------------------

//Mapping for in-memory databases (it just uses malloc and realloc)
char ndb_mapMem(nDB* db) {
	assert(db && !db->map);	//There must be a database object, and its map must be empty
		
	if (!db->size) db->size = NDB_STARTSIZE;
	
	ndb_log(NDB_LMEM,"MAP: Allocating memory of size %lu\n",db->size);
	
	db->map = calloc(db->size,1);	//Allocates the memory amount, setting to 0
	
	if (db->map) 
		return 0;	//Allocation succeeded
	
	ndb_errSetSys(db);	//Sets that there was a system error
	return -1;
}
char ndb_mapMemResize(void* db, NDB_LOC size) {
	char* ptr;		//A temporary variable used to hold allocation
	
	assert(db && size && ((nDB*)db)->map);		//Both database and new size must exist
	assert(size > ((nDB*)db)->size);	//Resize can only expand
	
	ndb_log(NDB_LMEM,"MAP: Resizing map %lu->%lu\n",((nDB*)db)->size,size);
	
	ptr = realloc(((nDB*)db)->map,size);
	
	if (ptr) {
		//The reallocation succeeded
		
		//We now set the newly allocated stuff to 0
		memset(ptr+((nDB*)db)->size,0,size-((nDB*)db)->size);
		
		//Set the new database parameters
		((nDB*)db)->map = ptr;
		((nDB*)db)->size = size;
		
		return 0;	//It succeeded
	}
	
	ndb_errSetSys(db);	//Sets that there was a system error
	return -1;			//At this point, the database still works
}
void ndb_mapMemClose(void* db) {
	assert(db && ((nDB*)db)->map);
	
	ndb_log(NDB_LMEM,"MAP: Freeing memory and closing map\n");
	
	free(((nDB*)db)->map);		//Frees memory
	
	((nDB*)db)->map  = NULL;	//Resets the variables
	((nDB*)db)->size = 0;
}

//Memory mapping for file databases - LINUX
#ifdef __linux__

#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

char ndb_mapFile(nDB* db, char* filename) {
	long pgsize = ndb_getPageSize();	//Gets the memory page size
	struct stat s;
	
	assert(db && !db->map);
	
	ndb_log(NDB_LMEM,"MAP: Mapping file %s\n",(filename?filename:""));
	
	if (!db->file) {
		//There is no file opened in the database
		
		assert(filename);	//Since there is no file, there MUST be a filename
		
		if (!db->readonly) {
			//The file is not flagged as readonly, so open for read-write
			db->file = open(filename, O_RDWR | O_CREAT);
			
			if (db->file == -1) {
				//File opening failed in RW
				ndb_errSetSys(db);
				ndb_log(NDB_LMSG,"map: File RW open failed, opening readonly\n");
				db->readonly = 1;
			}
		}
		
		if (db->readonly) {
			//The file is to be opened readonly
			db->file = open(filename, O_RDONLY);
			if (db->file == -1) {
				//File opening failed
				ndb_errSetSys(db);
				ndb_log(NDB_LMEM,"map: Failed to open file, aborting.\n");
				return -1;
			}
		}
	} else { ndb_log(NDB_LMEM,"map: File already open, continuing\n"); }
	
	//Read the file statistics
	if (fstat(db->file,&s)==-1) {
		ndb_errSetSys(db);
		ndb_log(NDB_LMEM,"map: Failed to get file stats, aborting.\n");
		return -2;
	}
	//Check if the file is normal
	if (!S_ISREG(s.st_mode)) {
		ndb_errSet(db,1);	//It is a file error
		ndb_log(NDB_LMEM,"map: File is not regular, aborting.\n");
		return -3;
	}
	
	//If given a specific size, or if file is empty, set default size
	if (!db->size && !s.st_size) {
		db->size = NDB_STARTSIZE;
		ndb_log(NDB_LMEM,"map: Using StartSize for map\n");
	} else if (db->size < s.st_size) {
		db->size = s.st_size;
		ndb_log(NDB_LMEM,"map: Using file size for map\n");
	} else {ndb_log(NDB_LMEM,"map: Using preset size for map\n");}
	
	//We now align the size to pagesize
	if (db->size % pgsize) {
		ndb_log(NDB_LMEM,"map: Adding size for page fit %d->",db->size);
		db->size = (db->size/pgsize+1)*pgsize;
		ndb_slog(NDB_LMEM,"%d\n",db->size);
	}
	
	//We now resize the file to the new size
	if (db->size > s.st_size) {
		//File needs to be resized
		if (db->readonly) {
			//The file is readonly!
			ndb_log(NDB_LMEM,"fmap: file is readonly, cannot add size, limited to %d bytes\n",s.st_size);
			ndb_log(NDB_LTDO,"TODO: Figure out a way to map beyond the end of readonly file\n");
			db->size = s.st_size;
		} else {
			//We can write to the file - so let's do it!
			lseek(db->file,db->size-1,SEEK_SET);	// -1 because file starts at 0!
			write(db->file,"",1);					//Write a 0 at that point, resizing the file
			lseek(db->file,0,SEEK_SET);				//Go back to 0
		}
	}
	
	//The file is now ready to be mapped
	db->map = mmap(0,db->size,PROT_READ | PROT_WRITE, (db->readonly?MAP_PRIVATE:MAP_SHARED),db->file,0);
	
	if (!db->map) {
		ndb_errSetSys(db);
		return -4;
	}
	
	//Hooray! The map is ready!
	return 0;
}
char ndb_mapFileResize(void* db, NDB_LOC size) {
	long pgsize = ndb_getPageSize();
	void* rmap;
	
	assert(db && ((nDB*)db)->file);
	assert(size > ((nDB*)db)->size);
	
	ndb_log(NDB_LMEM,"MAP: Resizing map %lu->%lu\n",((nDB*)db)->size,size);
	
	//If the file is readonly, fail
	if (((nDB*)db)->readonly) {
		ndb_log(NDB_LMEM,"map: Resizing is disabled for readonly files\n");
		ndb_log(NDB_LTDO,"TODO: figure out how to resize mapping of readonly files\n");
		ndb_errSet(((nDB*)db),1);
		return 1;	//It can revive easily from the error - nothing was changed
	}
	
	//Align size to page-fit
	if (size % pgsize) {
		ndb_log(NDB_LMEM,"map: Adding size for page fit %d->",size);
		size = (size/pgsize+1)*pgsize;
		ndb_slog(NDB_LMEM,"%d\n",size);
	}
	
	lseek(((nDB*)db)->file,size-1,SEEK_SET);	//Writes a byte at the new size, resizing the file
	write(((nDB*)db)->file,"",1);
	
#ifdef _GNU_SOURCE
	rmap = mremap(((nDB*)db)->map,((nDB*)db)->size,size,MREMAP_MAYMOVE);
#else
	//TODO: An alternate way to remap, without using mremap
	#error mremap unavailable - define _GNU_SOURCE!
#endif
	
	if (!rmap) {
		ndb_errSetSys((nDB*)db);
		return -1;					//The database is still intact if resize failed
	}
	
	((nDB*)db)->map = rmap;		//Write the new values to the database
	((nDB*)db)->size = size;
	
	return 0;		//It succeeded!
}
void ndb_mapFileClose(void* db) {
	assert(db);
	
	ndb_log(NDB_LMEM,"MAP: Closing file mapping\n");
	
	//Unmap and close the file
	if (((nDB*)db)->map)	munmap(((nDB*)db)->map,((nDB*)db)->size);
	if (((nDB*)db)->file)	close(((nDB*)db)->file);
	
	//Clears the file stuffs
	((nDB*)db)->readonly=0;
	((nDB*)db)->file = 0;
	((nDB*)db)->map=NULL;
	((nDB*)db)->size = 0;
}
#endif

//Memory mapping for file databases - WINDOWS
#ifdef _WIN32

//Windows.h was already included in ndb_def.h

char ndb_mapFile(nDB* db, char* filename) {
	long pgsize = ndb_getPageSize();
	LARGE_INTEGER filesize;
	
	assert(db && !db->map);
	
	ndb_log(NDB_LMEM,"MAP: Mapping file %s\n",(filename?filename:""));
	
	if (!db->file || db->file == INVALID_HANDLE_VALUE) {
		assert(filename);	//The filename must exist at this point
		
		if (!db->readonly) {
			db->file = CreateFile(filename,GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE,0,OPEN_ALWAYS,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,0);
			if (db->file == INVALID_HANDLE_VALUE) {
				//File opening failed in RW
				ndb_errSetSys(db);
				ndb_log(NDB_LMSG,"map: File RW open failed, opening readonly\n");
				db->readonly = 1;
			}
		}
		
		if (db->readonly) {
			db->file = CreateFile(filename,GENERIC_READ,
					FILE_SHARE_READ | FILE_SHARE_WRITE,0,OPEN_ALWAYS,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,0);
			if (db->file ==  INVALID_HANDLE_VALUE) {
				//File opening failed
				ndb_errSetSys(db);
				ndb_log(NDB_LMEM,"map: Failed to open file, aborting.\n");
				return -1;
			}
		}
	} else { ndb_log(NDB_LMEM,"map: File already open, continuing\n"); }
	
	//Read the file size
	if (!GetFileSizeEx(db->file,&filesize)) {
		ndb_errSetSys(db);
		ndb_log(NDB_LMEM,"map: Failed to get file stats, aborting.\n");
		return -2;
	}
	
	//If given a specific size, or if file is empty, set default size
	if (!db->size && !filesize.QuadPart) {
		db->size = NDB_STARTSIZE;
		ndb_log(NDB_LMEM,"map: Using StartSize for map\n");
	} else if (db->size < filesize.QuadPart) {
		db->size = filesize.QuadPart;
		ndb_log(NDB_LMEM,"map: Using file size for map\n");
	} else {ndb_log(NDB_LMEM,"map: Using preset size for map\n");}
	
	//We now align the size to pagesize
	if (db->size % pgsize) {
		ndb_log(NDB_LMEM,"map: Adding size for page fit %d->",db->size);
		db->size = (db->size/pgsize+1)*pgsize;
		ndb_slog(NDB_LMEM,"%d\n",db->size);
	}
	
	//We now resize the file to the new size
	if (db->size > filesize.QuadPart && db->readonly) {
		//The file is readonly!
		ndb_log(NDB_LMEM,"fmap: file is readonly, cannot add size, limited to %d bytes\n",filesize.QuadPart);
		ndb_log(NDB_LTDO,"TODO: Figure out a way to map beyond the end of readonly file\n");
		db->size = filesize.QuadPart;
	}
	
	//This is annoying, Msoft uses split 64 bit values
	filesize.QuadPart = db->size;
	
	//Create the file mapping
	db->fmap = CreateFileMapping(db->file,NULL,
		(db->readonly?PAGE_READONLY:PAGE_READWRITE),filesize.HighPart,filesize.LowPart,0);
	
	if (db->fmap == INVALID_HANDLE_VALUE) {
		ndb_errSetSys(db);
		return -4;
	}
	
	//Create the actual map
	db->map = MapViewOfFile(db->fmap,(db->readonly?FILE_MAP_READ:FILE_MAP_WRITE),0,0,0);
	
	if (!db->map) {
		ndb_errSetSys(db);
		return -4;
	}
	
	//Hooray! The map is ready!
	return 0;
}
char ndb_mapFileResize(void* db, NDB_LOC size) {
	long pgsize = ndb_getPageSize();	//Gets the system page size
	LARGE_INTEGER i;
	HANDLE h;
	
	assert(db && ((nDB*)db)->file && ((nDB*)db)->fmap && size);
	assert(size > ((nDB*)db)->size);
	
	ndb_log(NDB_LMEM,"MAP: Resizing map %lu->%lu\n",((nDB*)db)->size,size);
	
	//If the file is readonly, fail
	if (((nDB*)db)->readonly) {
		ndb_log(NDB_LMEM,"map: Resizing is disabled for readonly files\n");
		ndb_log(NDB_LTDO,"TODO: figure out how to resize mapping of readonly files\n");
		ndb_errSet(((nDB*)db),1);
		return 1;	//It can revive easily from the error - nothing was changed
	}
	
	//Align size to page-fit
	if (size % pgsize) {
		ndb_log(NDB_LMEM,"map: Adding size for page fit %d->",size);
		size = (size/pgsize+1)*pgsize;
		ndb_slog(NDB_LMEM,"%d\n",size);
	}
	
	//Windows, y u use weird structures?
	i.QuadPart = size;
	
	h = CreateFileMapping(((nDB*)db)->file,NULL,PAGE_READWRITE,i.HighPart,i.LowPart,0);
	
	if (h != INVALID_HANDLE_VALUE) {
		//Awww yea!
		char* tmp = (char*)MapViewOfFile(h,FILE_MAP_WRITE,0,0,0);
		if (tmp) {
			((nDB*)db)->map = tmp;
			CloseHandle(((nDB*)db)->fmap);
			((nDB*)db)->fmap = h;
			((nDB*)db)->size = size;
			return 0;
		}
	}
	ndb_errSetSys((nDB*)db);
	return -1;					//The database is still intact if resize failed
}
void ndb_mapFileClose(void* db) {
	assert(db);	//The stuff has to exist for this to work
	
	ndb_log(NDB_LMEM,"MAP: Closing file mapping\n");
	
	if (((nDB*)db)->fmap)	CloseHandle(((nDB*)db)->fmap);
	if (((nDB*)db)->file)	CloseHandle(((nDB*)db)->file);
	
	//Clears the file stuffs
	((nDB*)db)->readonly=0;
	((nDB*)db)->file = 0;
	((nDB*)db)->fmap = 0;
	((nDB*)db)->map=NULL;
	((nDB*)db)->size = 0;
}
#endif

//Gets the system page size
long ndb_getPageSize() {
#ifdef __linux__
	return sysconf(_SC_PAGESIZE);
#endif
#ifdef _WIN32
	SYSTEM_INFO sysinf;
	GetSystemInfo(&sysinf);
	return sysinf.dwPageSize;
#endif
}

//EOF

