/**************************************************************************************

                                  nDB Definitions

**************************************************************************************/
//Tab Width: 4 Character:\t

//Default Values for various parameters

//Pager defaults - these can be changed during runtime for each file, they are just
//	the default values

#ifndef NDB_PAGESIZE		//The default page size in bytes
#	define NDB_PAGESIZE 1024
#endif
#ifndef NDB_START			//The location in the file to start page-database
#	define NDB_START 20
#endif
#ifndef NDB_MAXCACHE		//Maximum amount of cache pages to store in RAM
#	define NDB_MAXCACHE 99999
#endif
#ifndef NDB_MAXALLOC		//The maximum amount of cache pages to preallocate
#	define NDB_MAXALLOC 30
#endif

//LOGGING:

#if defined(NDB_VERBOSE)	//If the database is to be verbose
//Level 1: Memory/File
#	define NDB_MVER 1	//Displays details every time memory is allocated/freed/memset
#	define NDB_FVER 1	//Displays details every time files are accessed using filesystem standard fncs

//Level 2: Page actions
#	define NDB_PLVER 2	//Displays details every time page is locked
#	define NDB_PGVER 2	//Displays details every time a page is read/written

//Level 3: Pager get
#	define NDB_PVER 3	//Displays page gets

//Level 4: New page/Discard page
#	define NDB_NPGVER 4	//Displays new page/Discard page requests
#	define NDB_TVER  4	//Displays details each time the trash structure is made/deleted

//Level 200: NDB Errors
#	define NDB_EVER 200	//Displays error every time an error is called
#endif

//#define NDBLOG_FLUSH	//Makes logs flush every time a message is written to them. Useful for making
						//	sure that stuff is written during debugging.

/*
This is the end of compile-time parameters
*/
//-------------------------------------------------------------------------------------

#ifndef NDB_DEF
#	define NDB_DEF

//The reason for this quite unusual structure is to allow the speedup of not
//	including any logging mechanisms in the database
#if defined(NDB_MVER) || defined(NDB_FVER) || defined(NDB_EVER) || defined(NDB_PVER)
	#define NDB_HASLOG 1	//It has a log
	//If a verbose function is defined, we give it a log file to write to
	#include "ndb_log.h"
#else
	#define NDB_HASLOG 0
	//To avoid compiler errors, define the functions as NULL
	#define ndb_getLogLevel() 0
	#define ndb_setLogLevel(l)

	#define ndb_getLogFile() NULL
	#define ndb_setLogFile(f) 

#endif

#define NDB_PAGEMINIMUM 30	//The minimum size of a page

/*
	These are the labels for different page types in the base database. These are what is 
	used to make sure all pages conform to specifications
*/
#define NDB_LTRASH	'T'
#define NDB_LFREE	'\0'
#define NDB_LHEAP	'H'

/*
	Inclusion of standard libraries n stuffs
*/
#include <stdlib.h>	//All of nDB uses the C standard library
#include <stdint.h>	//Includes fixed-width integers for platform-independence
#include <assert.h>	//Allows for checking program progress in debug

#ifdef _MSC_VER		//MSVC is very annoying for some things
#	undef inline
#	define inline __inline
#endif

//-------------------------------------------------------------------------------------
//Definitions

#define NDB_PGNUM uint64_t	//As of now, database page numbers are 64 bit unsigned ints
#define NDB_FLOC  uint64_t	//As of now, locations in the database are also 64 bit ints

//The database objects are just links to their page numbers
typedef NDB_PGNUM ndb_obj;
	//The database "objects" are really just pagenumbers, which once passed to the correct
	//	function are read and acted upon.


//
//The cache page structure - its use is the storage of pages being used.
typedef struct {
	uint8_t users;			//The number of places where the page is checked out
	uint8_t modified	:1;	//Boolean, whether the contents of the page have been 
							//	modified since loading
	uint8_t writing		:1;	//Boolean, whether a process has write lock
	uint8_t readers;		//Amount of reading threads
	uint8_t* data;			//The contents of the page
	
	NDB_PGNUM page;			//The page number of the page
} ndb_page;

//
//The Database Object. It contains all of the database's information.
typedef struct {
	//BOOLEANS
	uint8_t readonly	:1;	//Whether the file can be modified.
	uint8_t freeFlush	:1;	//Write to file each time modified page is freed
	
	/*	TODO: The structure locations might be better off on page 0
		TODO: The trash can save space by using an NDB_FLOC, instead of extra space
	*/
	//TRASH STRUCTURE LOCATION
	NDB_PGNUM trash;		//The location of the recycling structure, if any.
	//HEAP STRUCTURE LOCATION
	NDB_FLOC heap;			//The location of the end of the heap structure
	
	//DATABASE INFO
	NDB_PGNUM pagenum;		//The total amount of pages in the database	
	uint16_t pagesize;		//The size in bytes of a page
	NDB_FLOC pagestart;		//The start location of paging
	
	//ERROR
	uint8_t err;			//The error code of the database. Defaults to 0.
	
	//CACHE
	uint16_t ppool_size;	//The allocated maximum size of the pagepool
	uint16_t ppool_amt;		//Current location in pagepool
	ndb_page** ppool;		//The pagepool array
	
	NDB_PGNUM cache_size;	//The allocated size of the cache
	NDB_PGNUM cache_amt;	//The current cache size
	ndb_page** cache;		//The cache array
	NDB_PGNUM cache_max;	//The maximal amount of cache pages to allow (0 for no limit)
	uint32_t cache_addon;	//The amount of pages to add to the end of the array
	
	
	//FILESYSTEM
	void* file;				//The file object: Since it uses a virtual file system, it
							//	includes a generalized pointer.
			
	void (*f)(void*);							//Flush file
	void (*c)(void*);							//Close file
	int (*r)(uint64_t,uint32_t,void*,void*);	//Read file
	int (*w)(uint64_t,uint32_t,void*,void*);	//Write file
		//Since there can be arbitrary filesystems, it includes pointers to the file
		//	manipulation functions in the database object itself
	
} nDB;

#endif 
//EOF

