/**************************************************************************************

                                        Pager
                                     Definitions

**************************************************************************************/
//Tab Width: 4 Character:\t

#ifndef NDB_PAGER
#	define NDB_PAGER 1

#include "ndb_def.h"	//Includes nDB structures

/*
	The pager is the place where all storage manipulation takes place, it abstracts
	away the cache and the file system giving a unified interface that takes care of
	all of the data, so that the database structures don't have to worry about anything
	beyond writing to pages.
*/


uint8_t ndb_pagerInit(nDB* db);
void ndb_pagerClose(nDB* db);
	//These functions prepare the database for paging, and then freeing the pager and
	//	getting it cleanly to file

ndb_page* ndb_getPage_(nDB* db, NDB_PGNUM pg);
#define ndb_getPage ndb_getPage_	//Allows overloading with logger
	//The function used to obtain any and all pages!

void ndb_freePage_(ndb_page* p);
#define ndb_freePage ndb_freePage_
	//Notifies the pager that the page pointer is no longer in use, allowing it to be dereferenced
	//	While strictly speaking this is not necessary for small databases, and definitely unneeded
	//	for in-memory databases, freePage allows the cache to reallocate the cache page to a different
	//	part of the database if the cache maxes out

void ndb_pgFlush(nDB* db, ndb_page* p);	//Flushes the given page to a file
void ndb_flush_(nDB* db);
#define ndb_flush ndb_flush_
	//Writes all there is in the cache to the file - if there is no file, then does nothing
	//	this is also in the log-overloading

//-------------------------------------------------------------------------------------
//Page modification functions to make it easier to read and write to the pages
//	Please note that you must set a read/write lock EACH TIME you use  them!

//Returns a pointer straight to the data;
void* ndb_pgptr(ndb_page* p,uint16_t location);

//String access functions
char ndb_pgstrw_(ndb_page* p, uint16_t location, char* string);
char* ndb_pgstrr_(ndb_page* p, uint16_t location);
uint16_t ndb_pgstrlen(ndb_page* p, uint16_t location);

#define ndb_pgstrw ndb_pgstrw_
#define ndb_pgstrr ndb_pgstrr_

//Object write/read functions
char ndb_pgw_(ndb_page* p, uint16_t location, uint16_t numbytes, void* object);
char ndb_pgr_(ndb_page* p, uint16_t location, uint16_t numbytes, void* object);

#define ndb_pgw ndb_pgw_
#define ndb_pgr ndb_pgr_

//Character access functions
uint8_t ndb_pggetc_(ndb_page* p, uint16_t location);
char ndb_pgputc_(ndb_page* p, uint16_t location, uint8_t c);

#define ndb_pggetc ndb_pggetc_
#define ndb_pgputc ndb_pgputc_

//Memsets, etc
char ndb_pgmemset_(ndb_page* p, uint16_t location, int value, uint16_t numbytes);
char ndb_pgclear_(nDB* db, ndb_page* p);

#define ndb_pgmemset ndb_pgmemset_
#define ndb_pgclear ndb_pgclear_

//Definitions of simplifier functions:
char ndb_pgwu16_(ndb_page* p, uint16_t location, uint16_t i);
char ndb_pgwu32_(ndb_page* p, uint16_t location, uint32_t i);
char ndb_pgwu64_(ndb_page* p, uint16_t location, uint64_t i);

uint16_t ndb_pgru16_(ndb_page* p, uint16_t location);
uint32_t ndb_pgru32_(ndb_page* p, uint16_t location);
uint64_t ndb_pgru64_(ndb_page* p, uint16_t location);

#define ndb_pgwu16 ndb_pgwu16_
#define ndb_pgwu32 ndb_pgwu32_
#define ndb_pgwu64 ndb_pgwu64_

#define ndb_pgru16 ndb_pgru16_
#define ndb_pgru32 ndb_pgru32_
#define ndb_pgru64 ndb_pgru64_

//Writes the label for the page, locking and unlocking automatically
void ndb_putlP_(ndb_page* p,uint8_t lbl);
#define ndb_putlP ndb_putlP_

//Compares n bytes of the object with the corresponding location in the page
uint16_t ndb_pgCompare(ndb_page* p, uint16_t location, void* object, uint16_t size);

//Simplifier read/write macros
#define ndb_pglw(a,b,c,d) do {ndb_pgWriteLock(a); ndb_pgw(a,b,c,d); ndb_pgWriteUnlock(a);} while (0)
#define ndb_pglr(a,b,c,d) do {ndb_pgRead(a); ndb_pgr(a,b,c,d); ndb_pgDone(a);} while (0)


//------------------------------------------------------------------------------------
//Logging
#if defined(NDB_PVER) && NDB_HASLOG

//Redefine the functions to be logging versions
#undef ndb_getPage
#undef ndb_freePage
#undef ndb_flush

#define ndb_getPage ndb_getPage_l
#define ndb_freePage ndb_freePage_l
#define ndb_flush ndb_flush_l

//File/Line Macros
#define ndb_getPage_l(a,b) ndb_getPage__(a,b,__FILE__,__LINE__)
#define ndb_freePage_l(a) ndb_freePage__(a,__FILE__,__LINE__)
#define ndb_flush_l(a) ndb_flush__(a,__FILE__,__LINE__)

//The logging versions of the functions
ndb_page* ndb_getPage__(nDB* db, uint64_t pg,char* loc,int line);
void ndb_freePage__(ndb_page* p,char* loc,int line);
void ndb_flush__(nDB* db,char* loc,int line);

#endif

#if defined(NDB_PGVER) && NDB_HASLOG
//Switch of definitions to logging versions of functions (which call the originals
//	internally)

#undef ndb_pgstrw
#undef ndb_pgstrr
#undef ndb_pgw
#undef ndb_pgr
#undef ndb_pggetc
#undef ndb_pgputc
#undef ndb_pgmemset
#undef ndb_pgclear
#undef ndb_pgwu16
#undef ndb_pgwu32
#undef ndb_pgwu64
#undef ndb_pgru16
#undef ndb_pgru32
#undef ndb_pgru64
#undef ndb_putlP

#define ndb_pgstrw ndb_pgstrw_l
#define ndb_pgstrr ndb_pgstrr_l
#define ndb_pgw ndb_pgw_l
#define ndb_pgr ndb_pgr_l
#define ndb_pggetc ndb_pggetc_l
#define ndb_pgputc ndb_pgputc_l
#define ndb_pgmemset ndb_pgmemset_l
#define ndb_pgclear ndb_pgclear_l
#define ndb_pgwu16 ndb_pgwu16_l
#define ndb_pgwu32 ndb_pgwu32_l
#define ndb_pgwu64 ndb_pgwu64_l
#define ndb_pgru16 ndb_pgru16_l
#define ndb_pgru32 ndb_pgru32_l
#define ndb_pgru64 ndb_pgru64_l
#define ndb_putlP ndb_putlP_l

//File/Line adder macros
#define ndb_pgstrw_l(a,b,c) ndb_pgstrw__(a,b,c,__FILE__,__LINE__)
#define ndb_pgstrr_l(a,b) ndb_pgstrr__(a,b,__FILE__,__LINE__)
#define ndb_pgw_l(a,b,c,d) ndb_pgw__(a,b,c,d,__FILE__,__LINE__)
#define ndb_pgr_l(a,b,c,d) ndb_pgr__(a,b,c,d,__FILE__,__LINE__)
#define ndb_pggetc_l(a,b) ndb_pggetc__(a,b,__FILE__,__LINE__)
#define ndb_pgputc_l(a,b,c) ndb_pgputc__(a,b,c,__FILE__,__LINE__)
#define ndb_pgmemset_l(a,b,c,d) ndb_pgmemset__(a,b,c,d,__FILE__,__LINE__)
#define ndb_pgclear_l(a,b) ndb_pgclear__(a,b,__FILE__,__LINE__)
#define ndb_pgwu16_l(a,b,c) ndb_pgwu16__(a,b,c,__FILE__,__LINE__)
#define ndb_pgwu32_l(a,b,c) ndb_pgwu32__(a,b,c,__FILE__,__LINE__)
#define ndb_pgwu64_l(a,b,c) ndb_pgwu64__(a,b,c,__FILE__,__LINE__)
#define ndb_pgru16_l(a,b) ndb_pgru16__(a,b,__FILE__,__LINE__)
#define ndb_pgru32_l(a,b) ndb_pgru32__(a,b,__FILE__,__LINE__)
#define ndb_pgru64_l(a,b) ndb_pgru64__(a,b,__FILE__,__LINE__)
#define ndb_putlP_l(a,b) ndb_putlP__(a,b,__FILE__,__LINE__)

//Definition of logger functions
char ndb_pgstrw__(ndb_page* p, uint16_t location, char* string,char* loc,int line);
char* ndb_pgstrr__(ndb_page* p, uint16_t location,char* loc,int line);
char ndb_pgw__(ndb_page* p, uint16_t location, uint16_t numbytes, void* object,char* loc,int line);
char ndb_pgr__(ndb_page* p, uint16_t location, uint16_t numbytes, void* object,char* loc,int line);
uint8_t ndb_pggetc__(ndb_page* p, uint16_t location,char* loc,int line);
char ndb_pgputc__(ndb_page* p, uint16_t location, uint8_t c,char* loc,int line);
char ndb_pgmemset__(ndb_page* p, uint16_t location, int value, uint16_t numbytes,char* loc,int line);
char ndb_pgclear__(nDB* db, ndb_page* p,char* loc,int line);
char ndb_pgwu16__(ndb_page* p, uint16_t location, uint16_t i,char* loc,int line);
char ndb_pgwu32__(ndb_page* p, uint16_t location, uint32_t i,char* loc,int line);
char ndb_pgwu64__(ndb_page* p, uint16_t location, uint64_t i,char* loc,int line);
uint16_t ndb_pgru16__(ndb_page* p, uint16_t location,char* loc,int line);
uint32_t ndb_pgru32__(ndb_page* p, uint16_t location,char* loc,int line);
uint64_t ndb_pgru64__(ndb_page* p, uint16_t location,char* loc,int line);
void ndb_putlP__(ndb_page* p,uint8_t lbl,char* loc,int line);

#endif

//-------------------------------------------------------------------------------------
//Page modifiers are in the cache!
#include "cache.h"

#endif
//EOF

