/**************************************************************************************

                                 Core of the Database
                                   Implementation

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "core.h"

#include "mapper.h"		//File mapping functions
#include "offsets.h"	//Header offsets

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//-------------------------------------------------------------------------------------

nDB* ndb_open(char* file) {
	nDB* db = malloc(sizeof(nDB));		//Allocates memory for the database
	
	//If opening had no problem, return the database
	if (!ndb_open_v2(db,file,0)) return db;
	
	//Failed, so free the memory
	free(db);
	return NULL;
}
char ndb_open_v2(nDB* db, char* file, uint8_t options) {
	assert(db);	//There has to be a database, no?
	
	ndb_log(NDB_LCOR,"OPEN: Opening database\n");
	
	memset(db,0,sizeof(nDB));	//Clears the database;
	
	//Check if readonly
	if (options & NDB_READONLY) db->readonly = 1;
	
	if (file) {
		//A file name was passed! We open a file-based database
		
		char chk = ndb_checkFile(file);
		
		if (chk == 2) db->readonly = 1;
		
		if (chk >= 0) {
			//The database can be opened
			
			if (!ndb_mapFile(db,file)) {
				//The file was mapped! So we now set some shit
				
				db->resize = ndb_mapFileResize;
				db->close = ndb_mapFileClose;
				
				if (chk == 1) {
					//The file did not exist yet, so write the header
					ndb_hwrite(db);
				}
				return 0;	//All went well
			}
		}
		
	} else {
		//The database is to be in-memory
		
		if (!ndb_mapMem(db)) {
			//The file was mapped successfully
			
			db->resize = ndb_mapMemResize;
			db->close = ndb_mapMemClose;
			
			//Write header
			ndb_hwrite(db);
			
			return 0;
		}
	}
	
	ndb_log(NDB_LMSG,"open: Failed to open database\n");
	return -1;
}

void ndb_close(nDB* db) {
	assert(db);
	ndb_close_v2(db);
	free(db);
}
void ndb_close_v2(nDB* db) {
	assert(db);
	
	ndb_log(NDB_LMSG,"CLOSE: Closing database\n");
	
	db->close((void*)db);	//Closes the memory map of the database
}

//Creates an identical copy of the database either to file or to memory
nDB* ndb_copy(nDB* db, char* filename) {
	nDB* newdb;
	
	assert(db);	//The old database has to exist, no?
	
	ndb_log(NDB_LCOR,"COPY: Copying database\n");
	
	if (filename && ndb_checkFile(filename) != 1) {
		ndb_log(NDB_LCOR,"copy: A file already exists in that location, aborting!\n");
		return NULL;
	}
	newdb = ndb_open(filename);	//Creates a new database
	
	if (newdb) {
		//The new database was created
		if (db->size > newdb->size) {
			ndb_log(NDB_LCOR,"copy: resizing new database to fit new contents\n");
			newdb->resize(newdb,db->size);	//Resizes the new database
		}
		if (newdb->size >= db->size) {
			//The new size is large enough to fit the old
			
			//Copy the actual database
			ndb_log(NDB_LCOR,"copy: Copying database contents...");
			memcpy(newdb->map,db->map,ndb_ohEnd(db));
			ndb_slog(NDB_LCOR,"Done\n");
			
			//Aww yeah
			return newdb;
		}
		
		//If it got here, something f*cked up
		ndb_close(newdb);
		
		if (filename) {			//Remove the file we just created
			remove(filename);
			ndb_log(NDB_LCOR,"copy: Removing file\n");
		}
	}
	
	ndb_log(NDB_LCOR,"copy: failed");
	return NULL;
	
}

//-----------------------------------------------------------------------------------

uint8_t ndb_getComp() {
	//Gets the compile-time variable
	uint8_t c;
	
	c = sizeof(NDB_LOC);
	c = c << 4;
	c += sizeof(NDB_SIZE);
	
	return c;
}

//Checks if a file is a database
char ndb_checkFile(char* filename) {
	char id[5];		//Used for checking starting nDB
	uint8_t i = 0;	//Used for checking characters
	FILE* f;		//The file
	
	assert(filename);	//There has to be a filename passed
	
	ndb_log(NDB_LHDR,"FCHECK: Checking file '%s'\n",filename);
	
	if (!(f = fopen(filename,"r"))) {
		//If the file could not be opened for read, it probably doesn't exist
		ndb_log(NDB_LHDR,"fcheck: file does not exist yet\n");
		return 1;	//Not a database file or can't open
	}
	//The file was opened for read
	
	memset(id,'\0',5);	//Clears the ID string, for comparison
	
	fread(id,4,1,f);	//Reads the first 4 characters from the file
	
	if (strcmp(id,"nDB")) {
		//The first 4 characters are NOT "nDB\0"
		ndb_log(NDB_LHDR,"fcheck: File is not an nDB database\n");
		fclose(f);
		return -1;
	}
	//The file is an nDB database
	
	i = getc(f);	//Get the character after the nDB\0, which is the readversion
	
	if (i > NDB_VER || !i) {
		//The readversion is beyond that of our database, we can't read the file
		ndb_log(NDB_LHDR,"fcheck: Database version mismatch\n");
		fclose(f);
		return -2;
	}
	
	i = getc(f);	//Get the writeversion
	
	if (i > NDB_VER) {
		//The writeversion is beyond that of our database, we can only open in readonly
		ndb_log(NDB_LHDR,"fcheck: Write version mismatch - can only open readonly\n");
		fclose(f);
		return 2;
	} else if (!i) {
		//It was 0! We need to fail
		ndb_log(NDB_LHDR,"fcheck: Database corrupted! Aborting!\n");
		fclose(f);
		return -3;
	}
	
	i = getc(f);	//Now we check for compile-compatibility
	
	if (i != ndb_getComp()) {
		ndb_log(NDB_LHDR,"fcheck: Databases are compile-time incompatible! Aborting!\n");
		fclose(f);
		return -4;
	}
	
	fclose(f);
	return 0;
}

//Writes the default header to map
void ndb_hwrite(nDB* db) {
	assert(db && db->map);
	assert(db->size > NDB_HSIZE);
	
	ndb_log(NDB_LHDR,"HWRITE: Writing database header\n");
	
	strcpy(db->map,"nDB");			//Write the starting database identifier
	db->map[4] = NDB_VER;			//Write the database read version
	db->map[5] = NDB_VER;			//Write the database write version
	db->map[6] = ndb_getComp();		//Write the database compile-time identifier
	
	ndb_ohEnd(db) = NDB_HSIZE;		//The endpoint is right after the header
	ndb_ohTable(db) = 0;			//The table structure is non-existent
	ndb_ohTrash(db) = 0;			//There is no trash structure
	ndb_ohIndex(db) = 0;			//The database Index does not exist yet
}

//EOF

