/**************************************************************************************

                            Table Manipulation Functions

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "table.h"
#include "offsets.h"
#include "allocator.h"

#include <string.h>

//-------------------------------------------------------------------------------------

uint8_t ndb_tableMk(nDB* db, uint8_t fieldnum, char* name) {
	NDB_LOC tbl;
	uint8_t tn;
	
	assert(db && fieldnum);
	
	if (name) ndb_log(NDB_LTBL,"TABLE: Creating table '%s'\n",name);
	else ndb_log(NDB_LTBL,"TABLE: Creating table\n");


	//Make sure the table struct exists
	if (!ndb_ohTable(db)) {
		ndb_log(NDB_LTBL,"table: Creating table index\n");
		//The table structure does not exist yet, so we allocate it.
		//	Alloc is guaranteed to have it at 0
		tbl = ndb_alloc(db,NDB_TSIZE);
		//WTF - gcc did not allow setting one to the other directly, it gave me fail.
		//	I don't have time to check if this is just a bug in GCC, or if I have 
		//	to be insane to do macro-function left-hand setting. Or both.
		ndb_ohTable(db) = tbl;
	}
	
	//Check to see if the table exists
	if (ndb_tableFind(db,name)) {
		//A table with that name already exists!
		ndb_log(NDB_LTBL,"table: A table named '%s' already exists!\n",name);
		return 0;
	}

	tbl = ndb_ohTable(db);
	
	if (ndb_otNum(db,tbl)>=255) {
		ndb_log(NDB_LTBL,"table: Already maxed out table number, can't create more!\n");
		ndb_errSet(db,5);
		return 0;
	}
	
	//We allocate a structure for the table with enough size to fit the table definition
	ndb_otable(db,tbl,++ndb_otNum(db,tbl)) = ndb_alloc(db,NDB_TSHEADSIZE+fieldnum*NDB_TFIELDSIZE);
	
	//Iterate the table number
	tn = ndb_otNum(db,tbl);
	
	//We now go to the index itself
	tbl = ndb_otable(db,tbl,ndb_otNum(db,tbl));
	
	//Remember: The index is guaranteed to be NULL
	
	//If a name exists, we write it to the database
	if (name) {
		//Now - there are two options, first, is that the name is a long string, so we allocate
		//	The other option is that the string fits snugly in the link, so we put it there instead,
		//	and click the first boolean - the isHere flag
		if (strlen(name) < sizeof(NDB_LOC)) {
			//The string can fit inside the location, so we copy it there
			strcpy(ndb_otsNameS(db,tbl),name);
			//We now set the first flag, boolean 0
			ndb_otsBoolSet(db,tbl,0,1);	//The string DID fit
		} else {
			//The string won't fit in the location, so we allocate memory for it
			ndb_otsName(db,tbl) = ndb_alloc(db, strlen(name)+1);
			//Copy it to the memory location
			strcpy(ndb_pstr(db,ndb_otsName(db,tbl)),name);			
		}
	}
	
	//Now we set the fieldnum in the table
	ndb_otsNum(db,tbl) = fieldnum;
	
	//Return the table number;
	return tn;
}

//Sets a table's field with the given values
void ndb_tfieldSet(nDB* db, NDB_LOC tbl, uint8_t fieldnum, uint8_t type, char* name) {
	assert(db && tbl && type);
	
	ndb_log(NDB_LTBL,"TABLE: Setting field %u to type %u\n",fieldnum,type);

	if (name) {ndb_log(NDB_LTBL,"table: field name is '%s'\n",name);}

	if (ndb_otsNum(db,tbl) > fieldnum) {
		//The table has the right amount of fields
		tbl = ndb_ltsField(tbl,fieldnum);	//Gets the field location
		
		//First copy the field name
		if (name) {
			if (strlen(name) < sizeof(NDB_LOC)) {
				//The string can fit inside the location, so we copy it there
				strcpy(ndb_otfNameS(db,tbl),name);
				//Now we set the first flag, boolean 0
				ndb_otfBoolSet(db,tbl,0,1);
			} else {
				//The string won't fit in the location, se we allocate memory for it
				ndb_otfName(db,tbl) = ndb_alloc(db, strlen(name)+1);
				//Copy it to the location
				strcpy(ndb_pstr(db,ndb_otfName(db,tbl)),name);
			}
		}
		
		//Next, copy the field type
		ndb_otfType(db,tbl) = type;
	} else {
		ndb_log(NDB_LTBL,"table: table contains wrong field amount.\n");
		ndb_errSet(db,10);
	}
	
}

//Finds the table with the given name
uint8_t ndb_tableFind(nDB* db, char* name) {
	NDB_LOC t = ndb_ohTable(db);
	uint16_t i;
	assert(db && name && t);
	
	for (i=1; i < ndb_otNum(db,t)+1 && strcmp(name,ndb_tableName(db,ndb_otable(db,t,i)));i++);
	
	if (i < ndb_otNum(db,t)+1) return (uint8_t)i;
	return 0;
}

NDB_LOC ndb_tableLoc(nDB* db,uint8_t tbl) {
	assert(db);
	
	//Just in case
	if (!tbl || !ndb_ohTable(db)) return 0;
	
	return ndb_otable(db,ndb_ohTable(db),tbl);
	
}
//Finds the field with the given name in the table
uint8_t ndb_tfieldFind(nDB* db, NDB_LOC tbl, char* name) {
	uint16_t i;
	assert(db && tbl && name);
	
	for (i=0; i < ndb_otsNum(db,tbl) && strcmp(name,ndb_tfieldName(db,tbl,(uint8_t)i));i++);
	
	if (i < ndb_otsNum(db,tbl)) return (uint8_t)i;
	return 0;
}

//Gets the field name - this is unsafe, to use for more than one comparison, since
//	the memory can be reallocated at any moment - and changing the variables
//	is strictly forbidden!
char* ndb_tfieldName(nDB* db,NDB_LOC tbl,uint8_t field) {
	assert(db && tbl);
	if (ndb_otsNum(db,tbl) > field) {
		//There is the right field amount
		tbl = ndb_ltsField(tbl,field);	//Gets the field location
		if (ndb_otsBoolGet(db,tbl,0)) {
			//The table name is right within the location
			return ndb_otfNameS(db,tbl);
		} else {
			//The name is in an allocation unit, so return THAT as a string
			return ndb_pstr(db,ndb_otfName(db,tbl));
		}
	}
	ndb_log(NDB_LTBL,"table: table contains wrong field amount.\n");
	ndb_errSet(db,10);
	return NULL;
}
char* ndb_tableName(nDB* db, NDB_LOC tbl) {
	assert(db && tbl);
	if (ndb_otsBoolGet(db,tbl,0)) {
		//The table name is right within the location
		return ndb_otsNameS(db,tbl);
	} else {
		//The name is in an allocation unit, so return THAT as a string
		return ndb_pstr(db,ndb_otsName(db,tbl));
	}
}

uint8_t ndb_tfieldType(nDB* db, NDB_LOC tbl, uint8_t field) {
	assert(db && tbl);
	if (ndb_otsNum(db,tbl) > field) {
		//There is the right field amount
		tbl = ndb_ltsField(tbl,field);	//Gets the field location
		return ndb_otfType(db,tbl);
	}
	ndb_log(NDB_LTBL,"table: table contains wrong field amount.\n");
	ndb_errSet(db,10);
	return 0;
}

//EOF
