/**************************************************************************************

                                  nDB Filesystem
                           Database Memory Manipulation

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "filesystem.h"

//------------------------------------------------------------------------------------
//Memory mapping for in-memory databases (as in, just malloc)

#include <string.h>

char ndb_mapMem_(nDB* db, uint64_t size) {
	assert(db && size);	//The database has to be in existence
	assert(!db->map);	//The map shouldn't exist
	
	db->map = calloc(size,1);	//Allocates the memory amount, setting the memory to 0
	
	if (!db->map) {
		ndb_errSetdb(db,1);	//Malloc had an internal error
		return -1;			//Return -1 on failure
	}
	
	db->size = size;
	return 0;
}

char ndb_mapMemResize(void* db, uint64_t newsize){
	char* ptr;	//Variable
	assert(db && newsize);	//The database pointer has to exist, and size can't be 0
	assert(((nDB*)db)->map);		//The map has to exist
	assert(newsize > ((nDB*)db)->size);
	
	ndb_log(NDB_LMEM,"MRESIZE: Resizing map %lu->%lu",((nDB*)db)->size,newsize);
	
	ptr = realloc(((nDB*)db)->map,newsize);
	
	if (ptr) {
	
		//Now we set the newly allocated stuff to 0
		memset(ptr+((nDB*)db)->size,0,newsize-((nDB*)db)->size);
		
		//Set the database parameters
		((nDB*)db)->map = ptr;
		((nDB*)db)->size = newsize;
		
		//It succeeded
		return 0;
	}
	ndb_errSetdb(((nDB*)db),1);	//Malloc had an internal error - probably ran out of memory
	return -1;
}

void ndb_mapMemClose(void* db) {
	assert(db && ((nDB*)db)->map);
	ndb_log(NDB_LMEM,"MCLOSE: Closing map");
	
	free(((nDB*)db)->map);
	((nDB*)db)->map  = NULL;
	((nDB*)db)->size = 0;
}

#if NDB_LOG

char ndb_mapMem__(nDB* db, uint64_t size, NDB_LF) {
	ndb_ilog(NDB_LMEM,"%8.8s:%-4d| MEM: Allocating map of size %lu\n",f,l,size);
	return ndb_mapMem_(db,size);
}
/*
char ndb_mapMemResize__(void* db, uint64_t newsize, NDB_LF) {
	ndb_ilog(NDB_LMEM,"%8.8s:%-4d| MRESIZE: Resizing map %lu->%lu\n",f,l,((nDB*)db)->size,newsize);
	return ndb_mapMemResize_(db,newsize);
}
void ndb_mapMemClose__(void* db, NDB_LF) {
	ndb_ilog(NDB_LMEM,"%8.8s:%-4d| MCLOSE: Closing map\n",f,l);
	ndb_mapMemClose_(db);
}
*/
#endif

//------------------------------------------------------------------------------------
//Memory mapping for file databases - system specific code!

#ifdef __linux__	//The linux/unix mmap code

#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

long ndb_getPageSize() {
	return sysconf(_SC_PAGESIZE);
}
char ndb_mapFile_(nDB* db, char* filename, uint8_t options) {
	long pgsize = ndb_getPageSize();	//Gets the memory page size
	struct stat s;
	
	assert(db);
	
	//Only the database actually has to exist - if there is no filename,
	//	then the file is already opened using SYSTEM!! routines (not stdio)
	
	if (options & NDB_FCLEAR) {
		db->readonly = 0;
		
		//Clears the previous files
		if (db->file) ndb_mapFileClose(db);
		if (db->map) ndb_mapMemClose(db); 
	}
	
	//If there is no file, open it!
	if (!db->file) {
		int f=0;	//The file descriptor
		assert(filename);	//If a file is not opened yet, a filename must exist!
		
		//Open the file if it is for read & write
		if (options & NDB_FRW && !db->readonly) {
			f = open(filename, O_RDWR | O_CREAT);
			if (f!=-1) {
				//File opening succeeded
				db->file = f;
				
			} else {
				//File opening failed
				ndb_log(NDB_LMSG,"mapFile: File RW open failed, opening for Readonly");
				db->readonly=1;	//Sets the readonly flag
			}
		} else {db->readonly = 1;}
		
		if (db->readonly) {
			f = open(filename, O_RDONLY);
			if (f==-1) {
				ndb_errSetdb(db,NDB_EFILE);
				return -2;
			} else {
				//Set the file descriptor
				db->file = f;
			}
		}
	} else ndb_log(NDB_LMEM,"fmap: File already opened, not checking for filename");
	
	//Read file statistics
	if (fstat(db->file,&s)==-1) {
			ndb_errSet(db);
			return -3;
	}
	if (!S_ISREG(s.st_mode)) {	//It is not a regular file!
		ndb_errSetdb(db,NDB_EFILE);
		return -4;
	}
	//If not given a size to start with, or the file is already larger than size
	//	then use the file's original size
	if (!db->size || s.st_size > db->size) {
		db->size = s.st_size;
		ndb_log(NDB_LMEM,"fmap: Using file size for map");
	} else ndb_log(NDB_LMEM,"fmap: Using preset size for map");
	
	//Now we align the size to a pagesize - because why not?
	if (db->size % pgsize) {	//If it does not divide evenly
		ndb_log(NDB_LMEM,"fmap: Adding size for page-fit %d->%d",db->size,(db->size/pgsize+1)*pgsize);
		db->size = (db->size/pgsize+1)*pgsize;
	}
	
	//Now we write 0s for the whole area beyond the end of the file
	if (db->size > s.st_size) {
		if (!db->readonly) {
			lseek(db->file,db->size-1,SEEK_SET);	//The -1 is there because file starts at 0
			write(db->file,"",1);				//The 0 is written at that point, so mapping will work
			lseek(db->file,0,SEEK_SET);
		} else {
			ndb_log(NDB_LMEM,"fmap: file is readonly, cannot add size, limited to %d bytes",s.st_size);
			ndb_log(NDB_LTODO,"TODO: Figure out a way to map beyond the end of readonly file");
			db->size = s.st_size;
		}
	}
	
	db->map = mmap(0,db->size,PROT_READ | PROT_WRITE,(db->readonly?MAP_PRIVATE:MAP_SHARED),db->file,0);
	
	if (!db->map) {
		ndb_errSetdb(db,NDB_EFILE);
		return -5;
	}
	
	return 0;
}

char ndb_mapFileResize(void* db, uint64_t newsize) {
	long pgsize = ndb_getPageSize();	//Gets the memory page size
	void* rmap;
	assert(db && ((nDB*)db)->file && newsize && newsize > ((nDB*)db)->size);	//Shit's gotta exist
	
	ndb_log(NDB_LMEM,"FRESIZE: Resizing map %lu->%lu",((nDB*)db)->size,newsize);
	
	if (newsize % pgsize) {	//Resize the allocation to an integer multiple of page
		ndb_log(NDB_LMEM,"fresize: Adding size for page-fit %d->%d",newsize,(newsize/pgsize+1)*pgsize);
		newsize = (newsize/pgsize+1)*pgsize;
	}
	
	if (((nDB*)db)->readonly) {
		ndb_log(NDB_LTODO,"TODO: figure out how to resize mapping of readonly files");
		ndb_errSetcustom(((nDB*)db),"fresize: Resizing is disabled for read-only files");
		return 1;	//It can revive easily from the error - nothing was changed
	}
	
	lseek(((nDB*)db)->file,newsize-1,SEEK_SET);
	write(((nDB*)db)->file,"",1);	//Sizes the file to be larger
	
	
#ifdef _GNU_SOURCE
	rmap = mremap(((nDB*)db)->map,((nDB*)db)->size,newsize,MREMAP_MAYMOVE);
#else
	//TODO: An alternate way to remap, without using mremap
	#error mremap unavailable - define _GNU_SOURCE!
#endif
	if (!rmap) {
		ndb_errSet(((nDB*)db));
		return -1;
	}
	
	((nDB*)db)->map = rmap;
	((nDB*)db)->size = newsize;
	
	return 0;
}
void ndb_mapFileClose(void* db) {
	assert(db);	//The stuff has to exist for this to work
	
	ndb_log(NDB_LMEM,"FCLOSE: Closing file");
	
	if (((nDB*)db)->map)	munmap(((nDB*)db)->map,((nDB*)db)->size);	//Unmap the file
	if (((nDB*)db)->file)	close(((nDB*)db)->file);	//Close the file - if anys
	
	//Clears the file stuffs
	((nDB*)db)->readonly=0;
	((nDB*)db)->file = 0;
	((nDB*)db)->map=NULL;
	((nDB*)db)->size = 0;
	
}

#endif

#ifdef _WIN32		//The windows mmap code

#include <Windows.h>

long ndb_getPageSize() {
	//Get the page size
	SYSTEM_INFO sysinf;
	GetSystemInfo(&sysinf);
	return sysinf.dwPageSize;
}

char ndb_mapFile_(nDB* db, char* filename, uint8_t options) {
	long pgsize = ndb_getPageSize();
	LARGE_INTEGER filesize;


	if (options & NDB_FCLEAR) {
		db->readonly = 0;

		//TODO 
		//Clears the previous files
		//if (db->file) ndb_mapFileClose(db);
		//if (db->map) ndb_mapMemClose(db);
	}

	//If there is no file, we open it
	if (!db->file) {
		HANDLE f = 0;	//The file handle
		assert(filename);	//If file is not opened yet, a filename must exist

		//Open the file if it is for RW
		if (options & NDB_FRW && !db->readonly) {
			f = CreateFile(filename,GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE,0,OPEN_ALWAYS,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,0);
			if (f !=  INVALID_HANDLE_VALUE) {
				//File open success
				db->file = f;
			} else {
				ndb_log(NDB_LMSG,"mapFile: File RW open failed, opening for Readonly");
				db->readonly = 1;
			}
		} else db->readonly =1;
		if (db->readonly) {
			f = CreateFile(filename,GENERIC_READ,
					FILE_SHARE_READ | FILE_SHARE_WRITE,0,OPEN_ALWAYS,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,0);
			if (f ==  INVALID_HANDLE_VALUE) {
				ndb_errSetdb(db,NDB_EFILE);
				return -2;
			} else db->file = f;	//Set the file handle
		} 
	} else ndb_log(NDB_LMEM,"fmap: File already opened, not checking for filename");

	//Obtain the file size
	
	if (!GetFileSizeEx(db->file,&filesize)) {
		ndb_errSet(db);
		return -3;
	}

	//If not given a size to start with, or the file is already larger than size
	//	then use the file's original size
	if (!db->size || filesize.QuadPart > db->size) {
		db->size = filesize.QuadPart;
		ndb_log(NDB_LMEM,"fmap: Using file size for map");
	} else ndb_log(NDB_LMEM,"fmap: Using preset size for map");
	
	//Now we align to pagesize
	if (db->size % pgsize) {	//If it does not divide evenly
		ndb_log(NDB_LMEM,"fmap: Adding size for page-fit %d->%d",db->size,(db->size/pgsize+1)*pgsize);
		db->size = (db->size/pgsize+1)*pgsize;
	}

	//If it is readonly, we make sure that we don't go beyond the file:
	if (db->readonly && filesize.QuadPart < db->size) {
		ndb_log(NDB_LMEM,"fmap: file is readonly, cannot add size, limited to %d bytes",filesize.QuadPart);
		ndb_log(NDB_LTODO,"TODO: Figure out a way to map beyond the end of readonly file");
		db->size = filesize.QuadPart;
	}

	if (db->readonly) {
		db->fmap = CreateFileMapping(db->file,NULL,PAGE_READONLY,0,0,0);
		if (db->fmap == INVALID_HANDLE_VALUE) {
			ndb_errSetdb(db,NDB_EFILE);
			return -6;
		}
		db->map = MapViewOfFile(db->fmap,FILE_MAP_READ,0,0,0);
	} else {
		LARGE_INTEGER i;
		i.QuadPart = db->size;
		db->fmap = CreateFileMapping(db->file,NULL,PAGE_READWRITE,i.HighPart,i.LowPart,0);
		if (db->fmap == INVALID_HANDLE_VALUE) {
			ndb_errSetdb(db,NDB_EFILE);
			return -6;
		}
		db->map = (char*)MapViewOfFile(db->fmap,FILE_MAP_WRITE,0,0,0);
	}	

	if (!db->map) {
		ndb_errSetdb(db,NDB_EFILE);
		return -5;
	}
	return 0;
}
char ndb_mapFileResize(void* db, uint64_t newsize) {
	LARGE_INTEGER i;
	HANDLE h;
	long pgsize = ndb_getPageSize();	//Gets the memory page size
	assert(db && ((nDB*)db)->file && ((nDB*)db)->fmap && newsize && newsize > ((nDB*)db)->size);	//Shit's gotta exist
	
	ndb_log(NDB_LMEM,"FRESIZE: Resizing map %lu->%lu",((nDB*)db)->size,newsize);
	
	if (newsize % pgsize) {	//Resize the allocation to an integer multiple of page
		ndb_log(NDB_LMEM,"fresize: Adding size for page-fit %d->%d",newsize,(newsize/pgsize+1)*pgsize);
		newsize = (newsize/pgsize+1)*pgsize;
	}
	
	if (((nDB*)db)->readonly) {
		ndb_log(NDB_LTODO,"TODO: figure out how to resize mapping of readonly files");
		ndb_errSetcustom(((nDB*)db),"fresize: Resizing is disabled for read-only files");
		return 1;	//It can revive easily from the error - nothing was changed
	}
	i.QuadPart = newsize;
	h = CreateFileMapping(((nDB*)db)->file,NULL,PAGE_READWRITE,i.HighPart,i.LowPart,0);
	if (h != INVALID_HANDLE_VALUE) {
		CloseHandle(((nDB*)db)->fmap);
		((nDB*)db)->fmap = h;
		((nDB*)db)->map = (char*)MapViewOfFile(((nDB*)db)->fmap,FILE_MAP_WRITE,0,0,0);
		((nDB*)db)->size = newsize;
		return 0;
	}
	
	ndb_errSetdb((nDB*)db,NDB_EFILE);
	return -1;
}
void ndb_mapFileClose(void* db) {
	assert(db);	//The stuff has to exist for this to work
	
	ndb_log(NDB_LMEM,"FCLOSE: Closing file");
	
	if (((nDB*)db)->fmap)	CloseHandle(((nDB*)db)->fmap);
	if (((nDB*)db)->file)	CloseHandle(((nDB*)db)->file);
	
	//Clears the file stuffs
	((nDB*)db)->readonly=0;
	((nDB*)db)->file = 0;
	((nDB*)db)->fmap = 0;
	((nDB*)db)->map=NULL;
	((nDB*)db)->size = 0;
}

#endif

#if NDB_LOG

char ndb_mapFile__(nDB* db, char* filename,uint8_t opt, NDB_LF) {
	ndb_ilog(NDB_LMEM,"%8.8s:%-4d| FMAP: Mapping file '%s'\n",f,l,filename);
	return ndb_mapFile_(db,filename,opt);
}
/*
char ndb_mapFileResize__(void* db, uint64_t newsize, NDB_LF) {
	ndb_ilog(NDB_LMEM,"%8.8s:%-4d| FRESIZE: Resizing map %lu->%lu\n",f,l,((nDB*)db)->size,newsize);
	return ndb_mapFileResize_(db,newsize);
}
void ndb_mapFileClose__(void* db, NDB_LF) {
	ndb_ilog(NDB_LMEM,"%8.8s:%-4d| FCLOSE: Closing file\n",f,l);
	ndb_mapFileClose_(db);
}
*/
#endif

//EOF

