/**************************************************************************************

                                  nDB Definitions

**************************************************************************************/
//Tab Width: 4 Character:\t

/*------------------------------------------------------------------------------------
                                 COMPILING NOTES
--------------------------------------------------------------------------------------

	If on linux, the file IO uses mmap and mremap, so it is required that you define 
	"_GNU_SOURCE" in your command-line parameters to gcc, since mremap is necessary.
	For some reason defining it in this header fails to work, so yeah:
	                    "gcc -D _GNU_SOURCE -Wall *.c"
	
	On windows, using Visual Studio 2010 it compiles without problem, although 
	at the time of writing, I am still unsure whether the windows version will work
	with large file sizes. 
	
	Yes, MSVC will give tons of warnings - they are irrelevant, compiling with GCC
	with all warnings enabled is still clean. Most warnings are due to comparisons,
	which are tested to work in both environments.

*/
/*------------------------------------------------------------------------------------
                             COMPILE-TIME PARAMETERS
--------------------------------------------------------------------------------------

	nDB's compile-time parameters are hard-wired into the database throughout its 
	runtime. While most variables won't affect the database's portability, making
	it slower at the worst, some (marked with a warning) will make the database 
	incapable of reading or writing files made by nDBs with different values.
	
	Before any variables, there is the 'DEBUG' qualifier, which disables all assertions
	used in the database - if this is commented out, then nDB does a debug build,
	if not, then it is doing a production build without assertions
	
	*///Whether not to debug
	//#define NDEBUG
	/*
	
	The first paramter is the resize variable, used in the database's structure allocator.
	When the allocator gets to the end of the database file or allocated region,
	it resizes the region - which depending on the database and OS can be a very
	expensive and time-consuming operation. The NDB_RESIZE variable tells the database
	by how many times the database's current size should it be resized after satisfying
	the current allocation. Notice that it CAN'T be less than 1
	
	*///The amount of allocations to resize by each time
	#define NDB_RESIZE 1.5f
	/*
	
	Now, there is the initial size of a newly created database, both in-memory and on file.
	This is a guideline, the mappers are under no obligation to make the size exactly that,
	in fact, the mappers will probably align to pagesize, but this variable gives a good
	estimate of how large a beginning database should be
	
	*///Database initial size
	#define NDB_STARTSIZE 1000
	/*
	
	Next are the "logging" variables. For debugging purposes, the database contains
	a logging mechanism, which can be either enabled or disabled at compile-time.
	The variable NDB_LOG does just that - if it is 1, then the logging mechanism is
	enabled, and if it is 0, then it is disabled. If it is commented out then
	it is assumed that it is disabled
	
	*///Whether or not the logging mechanism is enabled
	#define NDB_LOG 1
	/*
	
	The log can work in two ways: either it includes file and line number before each
	message for easier debugging, or it can skip them entirely. This definition tests
	whether to include file line number for log messages
	
	*///Whether or not to include file and line numbers in log
	#define NDB_LOGFLINE
	/*
	
	The next variable is a very quick definition: whether or not the log should be
	flushed after every message - if defined, then it will flush, if not, then it 
	won't. Keeping this on is more useful for debugging purposes, since it makes sure
	that it writes down every message before continuing. Just comment it out if it 
	is not wanted
	
	*///Whether or not to flush the logfile
	#define NDB_LOGFLUSH
	/*
	
	The logging mechanism makes use of what are called 'log levels'. There are 256 
	levels, from 0 to 255. Each level can be used multiple times. They are used to 
	separate low-level events from high-level ones, so that different amounts of
	data can be printed. Log level 0 is lowest - everything is printed, so every
	notification that the database makes is shown. Log level 255 is highest, meaning
	that nothing is shown. Different parts of the database have different levels, so 
	that it is easier to decipher what is going on.
	The log levels MUST be defined! The user can only change the INTEGER values.
	
	*///Log level definitions
	#define NDB_LERR	200		//Database errors
	#define NDB_LMSG	100		//Database messages (like "Opening for readonly")
	#define NDB_LTDO	0		//TODO statements
	#define NDB_LMEM	1		//Memory allocations and mappings
	#define NDB_LLOG	1		//Logger notifications, like "Changed log level"
	#define NDB_LALC	2		//Database internal allocator
	#define NDB_LHDR	2		//Database header writing
	#define NDB_LCOR	10		//Database opening & closing
	#define NDB_LSTR	5		//The database structure level
	#define NDB_LTBL	7		//The SQL-database level (table-level)
	#define NDB_LOBJ	7		//The Object level - so using the structures to do something
	/*
	
	WARNING: Beyond this point are variables that affect the database workings themselves,
		these are dangerous to change if you don't know what you're doing. The database
		does NOT check for variable overflows, the parameters here are assumed to be 
		optimally matched. Furthermore, these MUST be of the correct type, because the database
		relies on certain characteristics of unsigned vs signed, etc.
		Furthermore, nDBs compiled with different values for these variables are mutually
		INCOMPATIBLE, meaning that one can't even read the other!
	
	
	NOTE: stdint is included in the database, so all of those definitions are fair game
	
	The first variable is the database pointer variable - it is an unsigned integer that
	must be able to point to any byte in the whole database. By default, this is a 64 bit
	unsigned integer, to allow for large database sizes, but if compiled for 32 bit systems, 
	a 32 bit uint should work fine - but remember that to make database files mutually
	compatible, they must be compiled with the same variable in that spot. Note that any
	lower than 32 bit and the database will simply asplode before even being able to use,
	since many structures are many kilobytes in size.
	
	*///The ndb file location/pointer variable
	#define NDB_LOC	uint64_t
	/*
	
	Next on the list is a related variable, one which holds the 'size' of a structure,
	meaning that it represents the maximum amount of contiguous memory that it can
	allocate. This is also an unsigned integer, which cannot go below 16 bit, since the
	structures of the database themselves are many kilobytes in size.
	
	*///The ndb database object size variable
	#define NDB_SIZE uint32_t
	/*
	
--------------------------------------------------------------------------------------
                           END OF COMPILE-TIME PARAMETERS
------------------------------------------------------------------------------------*/

//These make sure the compile-time parameters are set
#ifndef NDB_LOG
#	define NDB_LOG 0
#endif

#if !defined(NDB_LOC) || !defined(NDB_SIZE)
#	error Location and Size variables were not defined correctly!
#endif

//-------------------------------------------------------------------------------------
//Begin the actual definitions

#ifndef NDB_DEF
#	define NDB_DEF

#define NDB_VER 1	//The nDB database version

//-------------------------------------------------------------------------------------
//Database Includes

#include <stdint.h>	//Fixed integer sizes for platform independence
#include <assert.h>	//The database code makes extensive use of assertions for debugging

//Windows has a funny way of needing extra stuff to do anything
#ifdef _WIN32
#	include <Windows.h>
#endif

//-------------------------------------------------------------------------------------
//Database Objects

//The nDB object
typedef struct {

	//The database access variables
	char* map;			//The database memory map, it is the way the file is accessed and
						//	modified. It must be a contiguous region of the size of the
						//	whole database
	NDB_LOC size;		//The total size of the memory map
	
#ifdef _WIN32			//Windows uses a different type for files, and it requires an
	HANDLE file;		//	extra variable to maintain a memory map
	HANDLE fmap;
#else
	int file;			//Linux/Unix uses an int as a file descriptor
#endif
	
	//Since there are arbitrary methods of creating a memory map, like mmap, or even
	//	just malloc, nDB allows any, as long as the functions conform to the style:
	//	The functions are passed the nDB object as a void pointer, and resize
	//	must resize the map and leave size with the new size, without destroying
	//	the old information, and making the new space all 0'd out, and close must
	//	close the file, all while resetting the file and size variables.
	
	char (*resize)(void*,NDB_LOC);	//This function resizes the map to the given size
	void (*close)(void*);			//This function closes the file and map
	
	uint8_t err;			//Error code - holds the database's error
	uint8_t readonly :1;	//Whether or not the database is readonly
	
	//Any database-specific information is included in the database itself
} nDB;

//The universal nDB iter struct, used as an iterator
typedef struct {
	nDB* db;		//The database object
	uint64_t data;	//Data that the iterator holds
	NDB_LOC loc;	//A link
} ndb_iter;

//-------------------------------------------------------------------------------------
//Logging & Error

#include "log.h"		//Database logging
#include "error.h"		//Database error

#endif
//EOF

