/**************************************************************************************

                       Pointer Offsets for Database Structures

**************************************************************************************/
//Tab Width: 4 Character:\t
/*
	Offsets are quite literally the database's core - each structure in the database has
	its own offsets which tell where each variable is supposed to be. So in reality, the
	structures are defined through their offsets, with each 'chunk' being an allocation unit.
	The magic of offsets is that they allow both left and right hand operations, due to 
	their macro-nature. 
	
	Prefixes:
	o: link to a variable
	l: location
*/

#ifndef NDB_OFFSETS
#	define NDB_OFFSETS

#include "ndb_def.h"

//These macros allow quick access to most locations, and are the basis of the offsets
#define	ndb_ptr(db,loc)		((void*)((db)->map+(loc)))
#define ndb_pstr(db,loc)	((char*)((db)->map+(loc)))
#define	ndb_ploc(db,loc)	(*(NDB_LOC*)((db)->map+(loc)))
#define	ndb_psize(db,loc)	(*(NDB_SIZE*)((db)->map+(loc)))
#define	ndb_pu8(db,loc)		(*(uint8_t*)((db)->map+(loc)))
#define	ndb_pu16(db,loc)	(*(uint16_t*)((db)->map+(loc)))
#define	ndb_pu32(db,loc)	(*(uint32_t*)((db)->map+(loc)))
#define	ndb_pu64(db,loc)	(*(uint64_t*)((db)->map+(loc)))
#define	ndb_p32(db,loc)		(*(int32_t*)((db)->map+(loc)))

//These functions allow multiple boolean storage in one byte
#define ndb_boolGet(db,loc,num) (ndb_pu8(db,loc) & (1<<(num)))	//Get boolean at bit num
void ndb_boolSet(nDB* db, NDB_LOC loc, uint8_t num, uint8_t b);	//Set boolean at bit num
void ndb_boolFlip(nDB* db, NDB_LOC loc, uint8_t num);			//Flip boolean at bit num

//-------------------------------------------------------------------------------------
//HEADER

#define NDB_HSIZE (7+4*sizeof(NDB_LOC))

#define ndb_ohEnd(db)		ndb_ploc(db,7)						//The database endpoint
#define ndb_ohTable(db)		ndb_ploc(db,7+sizeof(NDB_LOC))		//The structure where tables are stored
#define ndb_ohIndex(db)		ndb_ploc(db,7+2*sizeof(NDB_LOC))	//The RootList
#define ndb_ohTrash(db)		ndb_ploc(db,7+3*sizeof(NDB_LOC))	//There is no trash structure

//-------------------------------------------------------------------------------------
//TABLE INDEX
//The table index allows for the definition of 255 unique tables

#define NDB_TSIZE (255*sizeof(NDB_LOC)+sizeof(uint8_t))

#define ndb_otNum(db,i)		ndb_pu8(db,(i)+255*sizeof(NDB_LOC))	//The amount of tables in the database
#define ndb_otable(db,i,n)	ndb_ploc(db,i+((n)-1)*sizeof(NDB_LOC))	//Given db, location of tableIndex, and number, returns table pointer

//Size of the tablestruct's head variables
#define NDB_TSHEADSIZE (sizeof(uint8_t)*2+sizeof(NDB_LOC))		//The size of the table struct's header

//The tablestruct's head variables
//Boolean:
//	0: Whether or not name fit in Name variable
#define ndb_otsBoolGet		ndb_boolGet							//Table struct's booleans
#define ndb_otsBoolSet		ndb_boolSet
#define ndb_otsBoolFlip		ndb_boolFlip
#define ndb_otsNum(db,i)	ndb_pu8((db),(i)+sizeof(uint8_t))		//The table struct's number of fields
#define ndb_otsName(db,i)	ndb_ploc(db,(i)+sizeof(uint8_t)*2)	//The table's name
#define ndb_otsNameS(db,i)	ndb_pstr(db,(i)+sizeof(uint8_t)*2)	//The table's name as string

//Size of a field
#define NDB_TFIELDSIZE (sizeof(uint8_t)*2+sizeof(NDB_LOC))

//Location of a field n, given i as the TableStruct
#define ndb_ltsField(i,n) ((i)+sizeof(uint8_t)*2+sizeof(NDB_LOC)+(n)*NDB_TFIELDSIZE)

#define ndb_otfBoolGet		ndb_boolGet							//The table field's booleans
#define ndb_otfBoolSet		ndb_boolSet
#define ndb_otfBoolFlip		ndb_boolFlip
#define ndb_otfType(db,i)	ndb_pu8(db,(i)+sizeof(uint8_t))		//The Type of variable stored in field
#define ndb_otfName(db,i)	ndb_ploc(db,(i)+sizeof(uint8_t)*2)	//The field's name
#define ndb_otfNameS(db,i)	ndb_pstr(db,(i)+sizeof(uint8_t)*2)	//The field's name in string format

//-------------------------------------------------------------------------------------
//DATA - the structure within which rows of a table are held
//DATA defines 3 things, which must be at the first location:
//First, is boolean, next is the data, and last is the table number.
//While it has a boolean, it only reserves the 0th one, leaving the 1-7 booleans to the higher level
//	structures

//Size of the data structure
#define NDB_DATASIZE	(2+sizeof(NDB_LOC))
#define NDB_DATALOC(loc)		((loc)+1)

#define ndb_odBoolGet		ndb_boolGet							//Table struct's booleans
#define ndb_odBoolSet		ndb_boolSet
#define ndb_odBoolFlip		ndb_boolFlip
//Next is a location
#define ndb_odData(db,loc) ndb_ploc(db,NDB_DATALOC(loc))		//The location at which table data is stored
#define ndb_odTable(db,loc) ndb_pu8(db,(loc)+1+sizeof(NDB_LOC))	//The table number of the data
//The DATA structure is the basis of all structures which can store data, like node and link.
//DATA defines these two things, but it only uses boolean 0

//-------------------------------------------------------------------------------------
//LISTINDEX
//	An index that keeps ordered data (used for sequences)

#define NDB_LINDEXPGNUM 4	//The number of links per index page
#define NDB_LINDEXSIZE(lisize) (3+3*sizeof(NDB_LOC)+NDB_LINDEXPGNUM*(lisize))

#define ndb_oliID(db,loc)		ndb_pu8(db,loc)		//The IndexID - indexes must have them as first thing
#define ndb_oliSize(db,loc)		ndb_pu8(db,loc+1)	//the size of a list item (up to 256 bytes)
#define ndb_oliNum(db,loc)		ndb_pu8(db,loc+2)	//The number of links on the page
#define ndb_oliPrev(db,loc)		ndb_ploc(db,loc+3+NDB_LINDEXPGNUM*ndb_oliSize(db,loc))	//The previous link
#define ndb_oliNext(db,loc)		ndb_ploc(db,loc+3+NDB_LINDEXPGNUM*ndb_oliSize(db,loc)+sizeof(NDB_LOC))	//The next link
#define ndb_oliFirst(db,loc)	ndb_ploc(db,loc+3+NDB_LINDEXPGNUM*ndb_oliSize(db,loc)+2*sizeof(NDB_LOC))	//The first page
#define ndb_oliGet(db,loc,n)	((loc)+3+(n)*ndb_oliSize(db,loc))	//Gets the location of element n in the index

//-------------------------------------------------------------------------------------
//TREEINDEX
//	An index of links in a direct structure

#define NDB_TINDEXSIZE	(sizeof(NDB_LOC)*257+2)

#define ndb_otiID(db,loc)		ndb_pu8(db,loc)	//The index ID
#define ndb_otiGet(db,loc,n)	ndb_ploc(db,(loc)+2+(n)*sizeof(NDB_LOC))	//Gets the link
#define ndb_otiPrev(db,loc)		ndb_ploc(db,(loc)+2+256*sizeof(NDB_LOC))	//Gets the previous page
#define ndb_otiLnk(db,loc)		ndb_pu8(db,(loc)+1)	//Gets the associated link

//-------------------------------------------------------------------------------------
///RootList
//	The root structure
#define NDB_RLSNUM	2
#define NDB_RLSIZE	(sizeof(NDB_LOC)*(NDB_RLSNUM))

#define ndb_orlGet(db,loc,n)	ndb_ploc(db,(loc)+(n)*sizeof(NDB_LOC))	//Get the associated link

//-------------------------------------------------------------------------------------
//NODE- AND NODE LOCATIONS
//	Inherits from DATA structure, to allow storing Node properties in a table

#endif

//EOF

