/**************************************************************************************

                                        Cache
                                  The Implementation

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "cache.h"
#include "error.h"	//Allows for calling error codes

#include <string.h>	//memset <3

//-------------------------------------------------------------------------------------
//Internal Functions
//	These functions are necessary for the proper functioning of the cache, but are not
//	exposed further to the database. They are prefixed with cache_, which mean they are
//	internal cache functions

//
//Page functions
//

//This allocates a new page
ndb_page* cache_mkPg(nDB* db) {

	//ndb_page* p = calloc(sizeof(ndb_page),1);	//Allocate and set all to 0
	ndb_page* p = malloc(sizeof(ndb_page));	//It is set to 0 before giving

	assert(db);	//Assuming it was checked earlier, it can't be NULL
	
	if (p) {
		//p->data = calloc(db->pagesize,1);	//Allocates page data, and sets it to 0
		p->data = malloc(db->pagesize);	//It is set to 0 before giving

		if (p->data) return p;	//If everything allocated correctly, return
		
		//Since we got here, it failed, so fail gracefully
		free(p);	//We want to avoid a memory leak
	}
	
	//Memory allocation failed for the page
	ndb_seterr(db,NDBERR_MEMORY);
	return NULL;
}

//This frees a page
void cache_freePg(ndb_page* p) {
	assert(p);		//I think it makes sure that it is not NULL. If not, then just check.
	assert(p->data);//I am not sure if this can ever be NULL in an allocated page
	
	free(p->data);	//Actually free the page and its associated data.
	free(p);
}


//
//PagePool Functions
//

//This function refills the internal pagepool of the database
char cache_ppRefill(nDB* db) {
	assert(db && db->ppool);	//Assuming the non-NULLity was checked earlier
	
	//We refill the array
	while (db->ppool_amt < db->ppool_size)
		db->ppool[db->ppool_amt++] = cache_mkPg(db);
	return 0;
}

//This function initializes the pagepool
char cache_ppInit(nDB* db) {
	assert(db);			//Assuming non-NULL was checked earlier
	assert(!db->ppool);	//The ppool should not yet exist
	
	//Create the pagepool array - does not need to be 0
	db->ppool = malloc(sizeof(ndb_page*) * db->ppool_size);
	
	//make sure it allocated
	if (db->ppool) {
		if (!cache_ppRefill(db))
			return 0;
		//It failed somewhere, so fail gracefully
		//TODO: Free the ppool pages that were allocated!
		free(db->ppool);
		db->ppool = NULL;
	} else ndb_seterr(db,NDBERR_MEMORY); //refill has its own error
	return -1;
}
//This function closes the pagepool
void cache_ppClose(nDB* db) {
	assert(db && db->ppool);	//These should not be NULL
	
	while (db->ppool_amt >=1)	//Clear all allocated pages
		cache_freePg(db->ppool[--db->ppool_amt]);
	free(db->ppool);			//Frees the pagepool array
	db->ppool = NULL;
}

char cache_ppResize(nDB* db, uint16_t size) {
	ndb_page** pool;			//Create a temporary storage variable
	
	assert(db && db->ppool);	//The arguments shouldn't be NULL;
	assert(size);
	
	//If the new size is smaller than the currently allocated array, then
	//	free the extra pages
	while (db->ppool_amt >= size)
		cache_freePg(db->ppool[--db->ppool_amt]);
	
	//Reallocate the array to the new size
	pool = realloc(db->ppool,sizeof(ndb_page*) * size);
	
	//Check if the allocation was successful
	if (pool) {
		db->ppool = pool;
		db->ppool_size = size;
		return 0;
	}
	
	//Could not allocate necessary memory
	ndb_seterr(db,NDBERR_MEMORY);
	return -1;
}

//-------------------------------------------------------------------------------------
//External Functions
//	These functions are exposed to the outside world

char ndb_cacheInit(nDB* db) {
	assert(db);
	assert(!db->cache);
	
	//First, we create the pagepool
	if (!cache_ppInit(db)) {
		//If creation was successful
		
		//Make the cache array as large as the page amount plus some 
		//	extra space for added pages.
		db->cache_size = db->pagenum + db->cache_addon;
		
		assert(db->cache_size);		//The cache size can't be 0
		
		//Allocate the cache array
		db->cache = calloc(db->cache_size * sizeof(ndb_page*),1);
		
		if (db->cache) return 0;	//All is well
		
		cache_ppClose(db);	//Deallocate pagepool
		
		//Could not allocate necessary memory
		ndb_seterr(db,NDBERR_MEMORY);
	}
	return -1;	//There was a failure
}
void ndb_cacheClose(nDB* db) {
	assert(db && db->cache && db->ppool);	//These need to be allocated for it to work
	
	while (db->cache_size >=1)
		ndb_cacheDelPage(db,db->cache[--db->cache_size]);
			//Returns all of the pages to the pagepool
	
	free(db->cache);	//Frees the cache array
	db->cache = NULL;
	
	cache_ppClose(db);	//Frees the pagepool
}
	
ndb_page* ndb_cacheGet(nDB* db, NDB_PGNUM page) {
	assert(db && db->cache);
	
	if (db->cache_size > page)	//Can only have the page if it is within the bounds of array
		return db->cache[page];	//Return the cache page
	return NULL;				//Page was out of bounds, so NULL
}

ndb_page* ndb_cacheNewPage(nDB* db) {
	assert(db && db->ppool);
	
	//Make sure there are pages left in the pool
	if (db->ppool_amt == 0) cache_ppRefill(db);
	
	//Now, refill might have run out of memory, or the next couple pages might
	//	be NULL,  so it goes down the ppool and returns the first non-NULL instance,
	//	or if all fails, returns NULL itself.
	while (db->ppool_amt >= 1 && !db->ppool[--db->ppool_amt]);
	
	//Zero the page
	db->ppool[db->ppool_amt]->users=0;
	db->ppool[db->ppool_amt]->modified=0;
	db->ppool[db->ppool_amt]->writing=0;
	db->ppool[db->ppool_amt]->readers=0;
	db->ppool[db->ppool_amt]->page=0;
	
	//Data must be valid!
	assert(db->ppool[db->ppool_amt]->data);
	
	memset(db->ppool[db->ppool_amt]->data,0,db->pagesize);
	
	//Return it
	return db->ppool[db->ppool_amt];
}
void ndb_cacheDelPage(nDB* db, ndb_page* p) {
	assert(db && db->ppool);
	if (p) {	//The page can be NULL
		if (db->ppool_amt < db->ppool_size)
			db->ppool[db->ppool_amt++] = p;
		else
			cache_freePg(p);
	}
}

char ndb_cacheAddPage(nDB* db, ndb_page* p) {
	assert(db && db->cache && p);
	
	//When adding a page, we have to make sure that it is within the cache array.
	//	If not, we have to expand the array so that it does fit
	if (db->cache_size <= p->page) {
	
		uint64_t tmp = (p->page + 1 + db->cache_addon)*sizeof(ndb_page*);
		uint64_t tmp2 = tmp - db->cache_size*sizeof(ndb_page*);
		
		ndb_page** newcache = realloc(db->cache,tmp);
			//Has to have +1, since the page is at a 1 offset!
		
		if (!newcache) {
			//Could not allocate necessary memory
			ndb_seterr(db,NDBERR_MEMORY);
			
			return -1;
		}
		
		//Sets the new cache pointer
		db->cache = newcache;
		
		//Set the newly allocated entries to 0
		memset(&(db->cache[db->cache_size]),0,tmp2);
		
		//Sets the new cache size
		db->cache_size = p->page +1+ db->cache_addon;
	}
	
	//The page is not supposed to overwrite itself!
	//assert(!db->cache[p->page]);
	//This has bee replaced with a runcheck to make sure that multithreading
	//	does not generate memory leaks and crashes. It does not call an error,
	//	since the pager simply gives up the page in that case.
	if (db->cache[p->page]) return 1;
	
	db->cache[p->page] = p;	//Finally, set the pointer to the new page

	//Set the total size of the database if a new page is added beyond current size
	if (p->page >=db->pagenum) db->pagenum = p->page + 1;
	
	//Increase the cache amount
	db->cache_amt++;
	
	return 0;
}
void ndb_cacheRemovePage(nDB* db, ndb_page* p) {
	assert(db && db->cache && p);
	
	//We can't check to make sure there are no users to the cache, because it shouldn't
	//	care at shutdown
	
	if (db->cache_size > p->page) {	//Sets the page to NULL in the cache
		assert(db->cache[p->page]);	//It should not try to remove pages that are not in the cache!
		db->cache[p->page] = NULL;
		db->cache_amt--;			//Set the amount of cache pages lower
	}
	
	ndb_cacheDelPage(db,p);
}

void ndb_cacheClean(nDB* db) {
	ndb_page* p;
	ndb_cacheIter i=0;
	
	assert(db && db->cache);
	
	while ((p = ndb_cacheIterate(db,&i))) {
		if (!p->users && !ndb_pgIsModified(p)) {
			p->users++;			//We are using the page
			ndb_pgWriteLock(p);	//Just so a thread doesn't try reading while
								//	it is being deleted
			ndb_cacheRemovePage(db,p);
		}
	}
}


ndb_page* ndb_cacheIterate(nDB* db,ndb_cacheIter* i) {	
	assert(db && db->cache && i);
	
	//If it is beyond the last element, it is done
	if (*i >= db->cache_size) return NULL;
	
	//Search down the array for the next cache element
	while (*i < db->cache_size && db->cache[(*i)++] == NULL);
	
	//Return the current element - if at the end, return las element, so NULL.
	return db->cache[*i-1];
}

//These are the thread-savers.
//TODO: Someone who actually knows what they are doing should do this the way it
//	is supposed to be done. I am a noob at this.


uint8_t ndb_pgInUse(ndb_page* p) {
	assert(p);
	return (ndb_pgIsLocked(p) || ndb_pgReaders(p));
}

uint8_t ndb_pgReaders(ndb_page* p) {
	assert(p);
	return p->readers;
}
uint8_t ndb_pgIsLocked(ndb_page* p) {
	assert(p);
	return p->writing;
}
uint8_t ndb_pgIsModified(ndb_page* p) {
	assert(p);
	return p->modified;
}

void ndb_pgRead_(ndb_page* p) {
	assert(p);
	while (p->writing);	//Busywait while there is a writer
	p->readers++;
	
}
void ndb_pgDone_(ndb_page* p) {
	assert(p);
	p->readers--;
}

void ndb_pgSetWriteLock_(ndb_page* p) {
	assert(p);
	while (p->writing);	//Busywait while there is another writer
	p->writing=1;		//Set the write flag
	p->modified=1;		//The page is written, ergo it is modified
	
}
void ndb_pgWriteLock_(ndb_page* p) {
	assert(p);
	ndb_pgSetWriteLock_(p);
	while (p->readers);	//Wait while there are readers
}
void ndb_pgWriteUnlock_(ndb_page* p) {
	assert(p);
	p->writing=0;
}

//-----------------------------------------------------------------------------
//LOGGING thread-savers
#if defined(NDB_PLVER) && NDB_HASLOG

void ndb_pgRead__(ndb_page* p,char* fname,int lnum) {
	assert(p);
	ndb_log(NDB_PLVER,"\n(PRLK   Pg:%-7lu| U/MW/R:%2u/%i%i/%i | Locked:%i | %s:%d )",p->page,
		p->users,p->modified,p->writing,p->readers,ndb_pgIsLocked(p),fname,lnum);
	ndb_pgRead_(p);
}
void ndb_pgDone__(ndb_page* p,char* fname,int lnum) {
	assert(p);
	ndb_log(NDB_PLVER,"\n(PRULK  Pg:%-7lu| U/MW/R:%2u/%i%i/%i | Locked:%i | %s:%d )",p->page,
		p->users,p->modified,p->writing,p->readers,ndb_pgIsLocked(p),fname,lnum);
	ndb_pgDone_(p);
}
void ndb_pgSetWriteLock__(ndb_page* p,char* fname,int lnum) {
	assert(p);
	ndb_log(NDB_PLVER,"\n(PSWLK  Pg:%-7lu| U/MW/R:%2u/%i%i/%i | Locked:%i | %s:%d )",p->page,
		p->users,p->modified,p->writing,p->readers,ndb_pgIsLocked(p),fname,lnum);
	ndb_pgSetWriteLock_(p);
}
void ndb_pgWriteLock__(ndb_page* p,char* fname,int lnum) {
	assert(p);
	ndb_log(NDB_PLVER,"\n(PWLK   Pg:%-7lu| U/MW/R:%2u/%i%i/%i | Locked:%i | %s:%d )",p->page,
		p->users,p->modified,p->writing,p->readers,ndb_pgIsLocked(p),fname,lnum);
	ndb_pgWriteLock_(p);
}
void ndb_pgWriteUnlock__(ndb_page* p,char* fname,int lnum) {
	assert(p);
	ndb_log(NDB_PLVER,"\n(PWULK  Pg:%-7lu| U/MW/R:%2u/%i%i/%i | Locked:%i | %s:%d )",p->page,
		p->users,p->modified,p->writing,p->readers,ndb_pgIsLocked(p),fname,lnum);
	ndb_pgWriteUnlock_(p);
}

#endif
//EOF

