/**************************************************************************************

                                       Filesystem
                                      Main Header

**************************************************************************************/
//Tab Width: 4 Character:\t

#ifndef NDB_FILE
#	define NDB_FILE

#include "ndb_def.h"

/*
	The Filesystem is the implementation of database file access. There are multiple ways to
	obtain file access. For example, there is the standard C implementation, which uses the C
	library, which should compile perfectly on most computers, and then there's system-specific
	file access. Here, faster methods of attaining file data, like mmap can be used. Also, this
	allows the possibility to implement straight-to-disk file, bypassing the overhead of a
	filesystem altogether. The functions are as follows:
	
		(file object) open(file name, read/write)
		
		close(file object)
		
		flush(file object) --Makes sure data is written to disk
		
		(number of bytes written) write(location to write to, number of bytes to write,
					pointer to object from which to write, file object)
					
		(number of bytes read) read(location to read from, number of bytes to read,
					pointer to object to which to read, file object)
	
	
	TODO: System-specific file access. Right now, only the standard C method is implemented for
		maximum portability.
*/

//These are the default functions used to access the file. They are set based on the system
//	For now, only the ANSI C version of file access is available
//Please note that instead of the functions with _ at the end, use the defined ones without!
#define ndb_ofile_	ndb_ansiOpen
#define ndb_cfile_	ndb_ansiClose
#define ndb_wfile_	ndb_ansiWrite
#define ndb_rfile_	ndb_ansiRead
#define ndb_ffile_	ndb_ansiFlush

//The reason for the extra definition is to allow logging
#define ndb_ofile ndb_ofile_
#define ndb_cfile ndb_cfile_
#define ndb_wfile ndb_wfile_
#define ndb_rfile ndb_rfile_
#define ndb_ffile ndb_ffile_

//These specify the type of file access to attempt
#define NDB_FREAD	1
#define NDB_FWRITE	1<<1

//The ANSI C implementation - it should work on any platform
#ifndef NDB_NOANSI
void* ndb_ansiOpen(char* filename, uint8_t options);
void ndb_ansiClose(void* file);
void ndb_ansiFlush(void* file);
int ndb_ansiWrite(uint64_t loc, uint32_t numbytes, void* writeTo, void* file);
int ndb_ansiRead(uint64_t loc, uint32_t numbytes, void* writeFrom, void* file);
#endif

//Now for verbose file access
#if defined(NDB_FVER) && NDB_HASLOG

//Clears the default macros
#undef ndb_ofile
#undef ndb_cfile
#undef ndb_wfile
#undef ndb_rfile
#undef ndb_ffile

//Sets new macros
#define ndb_ofile(n,o) ndbver_ofile(n,o,__FILE__,__LINE__)
#define ndb_cfile(a) ndbver_cfile(a,__FILE__,__LINE__)
#define ndb_wfile(a,b,c,d) ndbver_wfile(a,b,c,d,__FILE__,__LINE__)
#define ndb_rfile(a,b,c,d) ndbver_rfile(a,b,c,d,__FILE__,__LINE__)
#define ndb_ffile(f) ndbver_ffile(f,__FILE__,__LINE__)

//These are the verbose replacement functions
void* ndbver_ofile(char* filename, uint8_t options,char* fname,int lnum);
void ndbver_cfile(void* file,char* fname,int lnum);
int ndbver_wfile(uint64_t loc, uint32_t numbytes, void* writeTo, void* file,char* fname,int lnum);
int ndbver_rfile(uint64_t loc, uint32_t numbytes, void* writeFrom, void* file,char* fname,int lnum);
void ndbver_ffile(void* file,char* fname,int lnum);

#endif

#endif
//EOF

