/**************************************************************************************

                                 nDB Page-Level Access

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "ndb_pg.h"
#include "error.h"
/*TRASH STRUCTURE:
	The trash is the simplest structure out there. It is a linked list of trash page
	arrays. The link that the database has to the trash page is the most recent (empty)
	trash structure page. The first 11 bytes are for the Trash label (1), link to next
	trash page (NDB_PGNSIZE), and trash page number (2). the rest of the page is used for an array
	of free pages. When a trash page is filled up, the next page it gets becomes the new 
	page, with its link pointing to the old page. Thus, a linked list of arrays is created.
	Since the 0th page is used as the core of the database, it is off-limits to the trash,
	which will automatically avoid it
*/
//-------------------------------------------------------------------------------------
//Label checking - for t3h easies

uint8_t ndb_checkLabelP(nDB* db, ndb_page* p, uint8_t lbl) {
	assert(p);
	if (lbl==ndb_pggetc(p,0)) return 0;
	ndb_seterr(db,NDBERR_CORRUPTED);
	return 1;
}
uint8_t ndb_checkLabel(nDB* db, NDB_PGNUM p, uint8_t lbl) {
	ndb_page* pg;
	uint8_t result;
	
	assert(db);	//Needs a database
	
	if (!(pg = ndb_getPage(db,p))) return 254;	//T3H MAJOR FAIL
	
	//Now we have the page
	ndb_pgRead(pg);
	result = ndb_checkLabelP(db,pg,lbl);
	ndb_pgDone(pg);
	ndb_freePage(pg);
	
	return result;
}

ndb_page* ndb_gcPage(nDB* db, NDB_PGNUM p, uint8_t lbl) {
	ndb_page* pg;
		
	assert(db);	//Needs a database
	
	if (!(pg = ndb_getPage(db,p))) return NULL;	//T3H MAJOR FAIL
	
	//Now we have the page
	ndb_pgRead(pg);
	
	if (!ndb_checkLabelP(db,pg,lbl))	//checkLabel returns 0 on success
		return pg;	//It is still reading! The user has to finish the readlock
	
	//Label matching failed
	ndb_pgDone(pg);
	ndb_freePage(pg);
	
	return NULL;
}

NDB_PGNUM ndb_locPage(nDB* db, NDB_FLOC p) {
	return ((p - db->pagestart)/db->pagesize);
}

uint16_t ndb_locLoc(nDB* db, NDB_FLOC p) {
	return ((p - db->pagestart)%db->pagesize);
}

//-------------------------------------------------------------------------------------
//Internal trash manipulation - they are prefaced with trash_ to make it apparent that
//	they are internal manipulation functions

//The start location for the *links* in the trash
#define TRASHLINKSTART (1 + NDB_PGNSIZE + 2)

void trash_add(nDB* db, NDB_PGNUM p) {
	ndb_page* pg;
	
	assert(db && db->trash);
	
	if ((pg=ndb_gcPage(db,db->trash,NDB_LTRASH))) {
		//The page was successfully obtained. If not, then it just
		//	drops it - won't do much harm.
		
		//Read the current link number
		uint16_t num = ndb_pgru16(pg,1+NDB_PGNSIZE);
		
		ndb_pgDone(pg);		//We are done reading
		
		//Now there are two options, if the link is within the page bounds, just add
		//	the link, and be done with it. If it is not, then do some magic, in other words,
		//	make the current page the new trash page
		if (num >= (db->pagesize - TRASHLINKSTART)/NDB_PGNSIZE) {
			//In this case, the page is full, so we make a new page
			ndb_freePage(pg);
			
			//The new page can be the one we just got!
			if ((pg = ndb_getPage(db,p))) {		//Only add to trash if the page exists
				ndb_pgWriteLock(pg);
				
				//ndb_pgclear(db,pg);				//Clears the page of its info
				ndb_pgwu16(pg,1+NDB_PGNSIZE,0);
					//Instead of clearing the whole page, it just clears this variable
				
				ndb_pgputc(pg,0,NDB_LTRASH);	//Writes the Trash page identifier
				ndb_pgwpg(pg,1,db->trash);		//Sets the current trash page to be last
				
				ndb_pgWriteUnlock(pg);			//End using the page
				ndb_freePage(pg);
				
				db->trash = p;					//The new trash location is this page
			}	
			//If not, then an error was already called
		} else {
			//In this case, there is room for at least one more page in the trash, so 
			//	we add it to the current page
			ndb_pgWriteLock(pg);
			
			//Write the page to the necessary location
			ndb_pgwpg(pg,TRASHLINKSTART+num*NDB_PGNSIZE,p);
			
			//Make the page one larger
			ndb_pgwu16(pg,1+NDB_PGNSIZE,num+1);
			
			//Unlock the page
			ndb_pgWriteUnlock(pg);
			
			//We are good to go
			ndb_freePage(pg);
		}
		
		
	}
	
	
}
NDB_PGNUM trash_get(nDB* db) {
	ndb_page* pg;
	NDB_PGNUM p;
	
	assert(db && db->trash);
	
	if ((pg=ndb_gcPage(db,db->trash,NDB_LTRASH))) {
		//Read the current link number
		uint16_t num = ndb_pgru16(pg,1+NDB_PGNSIZE);
		
		if (num) {	//There do exist links!
			
			//Read the page-link at the given location
			p = ndb_pgrpg(pg,TRASHLINKSTART+NDB_PGNSIZE*(--num));
			
			ndb_pgDone(pg);	//End readlock
			
			//Lower the link amount by 1
			ndb_pgWriteLock(pg);
			ndb_pgwu16(pg,1+NDB_PGNSIZE,num);
			ndb_pgWriteUnlock(pg);
			
			ndb_freePage(pg);	//Done using the page
			
			return p;	//Returns the read link page
		
		} else {	//The link number was 0
			//There are no links on this page, so move the trash back to the last page, if 
			//	such a page exists. If not, clean up and return 0;
			p=ndb_pgrpg(pg,1);	//Read the next page link			
			
			ndb_pgDone(pg);		//We are done with the page
			ndb_freePage(pg);
			
			if (p) {	//If there exists a link back
				NDB_PGNUM pnum = db->trash;
				db->trash = p;
				return pnum;
			}
			//There does not exist a link back, so we return a 0, meaning that it failed
		}
	}
	return 0;	//It failed to obtain a page
}

//-------------------------------------------------------------------------------------
//The functions that allow new pages and returning unused pages

NDB_PGNUM ndb_newPage_(nDB* db) {
	assert(db);
	if (db->trash) {
		NDB_PGNUM p = trash_get(db);
		if (p) return p;	//If the trash has a page to give
	}
	return db->pagenum++;	//Adds to the pagenum, and returns the page!
	
}
void ndb_delPageP_(nDB* db, ndb_page* p) {
	NDB_PGNUM pg;
	assert(db && p);
	pg = p->page;
	ndb_putlP(p,NDB_LFREE);	//Labels the page as free
	ndb_freePage_(p);		//Frees the page from the cache
	if (db->trash)
		trash_add(db,pg);	//If there is trash, add the page
}

//-------------------------------------------------------------------------------------
//Trash structure modification functions

void ndb_delTrash_(nDB* db) {
	ndb_page* pg;
	
	assert(db && db->trash);
	
	//Now we loop through all the trashes, setting labels to free everywhere we can
	while (db->trash) {
		if ((pg = ndb_gcPage(db,db->trash,NDB_LTRASH))) {
			//In the structure, the link to the next trash page is right after the label
			db->trash = ndb_pgrpg(pg,1);
			ndb_pgDone(pg);
			
			ndb_putlP(pg,NDB_LFREE);	//Labels the page as free
			
			ndb_freePage(pg);			//We are done with this page
			//And that's all that has to be done.
		} else {
			//A NULL page was returned - this is not supposed to happen. An error occurred,
			//	and an error was sent. All we need to do is return an error code
			db->trash = 0;
		}
		
	}
}
uint8_t ndb_mkTrash_(nDB* db) {
	NDB_PGNUM p;
	ndb_page* pg;
	
	assert(db && !db->trash);	//Make sure everything exists correctly
	
	//Gets the page
	pg = ndb_newPageP(db);
	if (!pg) return 1;
	
	p = pg->page;
	
	assert(p !=0);	//This should never happen
	
	
	//The locking is just a precaution - nobody should be using it yet anyways
	ndb_pgWriteLock(pg);
	ndb_pgclear(db,pg);				//Clears the page
	ndb_pgputc(pg,0,NDB_LTRASH);	//Writes the trash label
	ndb_pgWriteUnlock(pg);
	
	ndb_freePage(pg);	//We are done using the page
	
	db->trash=p;	//Set the trash location.
	
	return 0;
}
void ndb_pickupTrash_(nDB* db) {
	NDB_PGNUM p = 0;
	
	ndb_mkTrash_(db);
	while ((p=ndb_pgFindLabel(db,p,NDB_LFREE))) {	//Iterates through looking for free pages
		trash_add(db,p);	//For each free page, add to trash
	}
}

//-------------------------------------------------------------------------------------
//Iterating through the database page by page

NDB_PGNUM ndb_pgIterate(nDB* db, NDB_PGNUM loc) {
	if (loc < db->pagenum-1) return ++loc;	//Goes through all the database pages
	return 0;	//If done with the database, returns 0
}

NDB_PGNUM ndb_pgFind(nDB* db, NDB_PGNUM loc, uint16_t location, void* object, uint16_t size) {
	ndb_page* p;
	char c=1;
	while ((loc=ndb_pgIterate(db,loc))) {
		if ((p= ndb_getPage(db,loc))) {	//If it gets the page
			ndb_pgRead(p);
			c=ndb_pgCompare(p,location,object,size);
			ndb_pgDone(p);
			ndb_freePage(p);
			if (!c) return loc;
		}
	}
	return 0;
}

NDB_PGNUM ndb_pgFindChar(nDB* db, NDB_PGNUM loc, uint16_t location, uint8_t lbl) {
	ndb_page* p;
	char c=1;
	while ((loc=ndb_pgIterate(db,loc))) {
		if ((p= ndb_getPage(db,loc))) {	//If it gets the page
			ndb_pgRead(p);
			c=(lbl==ndb_pggetc(p,location));
			ndb_pgDone(p);
			ndb_freePage(p);
			if (c) return loc;
		}
	}
	return 0;
}

//-------------------------------------------------------------------------------------
//Next come the logging functions

#if NDB_HASLOG && defined (NDB_NPGVER)

NDB_PGNUM ndb_newPage__(nDB* db, char* fname, int lnum) {
	NDB_PGNUM p = ndb_newPage_(db);
	ndb_log(NDB_NPGVER,"\n(NEWPAGE Pg:%-6lu| %s:%i )",p,fname,lnum);
	return p;
}
void ndb_delPageP__(nDB* db, ndb_page* p, char* fname, int lnum) {
	ndb_log(NDB_NPGVER,"\n(DELPAGE Pg:%-6lu| U/MW/R:%2u/%i%i/%i | %s:%i )",p->page,
		p->users,p->modified,p->writing,p->readers,fname,lnum);
	ndb_delPageP_(db,p);
}

#endif

#if NDB_HASLOG && defined (NDB_TVER)

void ndb_delTrash__(nDB* db, char* fname, int lnum) {
	ndb_log(NDB_TVER,"\n(DELTRASH %s:%i )",fname,lnum);
	ndb_delTrash_(db);
}
uint8_t ndb_mkTrash__(nDB* db, char* fname, int lnum) {
	ndb_log(NDB_TVER,"\n(MKTRASH %s:%i )",fname,lnum);
	return ndb_mkTrash_(db);
}
void ndb_pickupTrash__(nDB* db, char* fname, int lnum) {
	ndb_log(NDB_TVER,"\n(REFILLTRASH %s:%i )",fname,lnum);
	ndb_pickupTrash_(db);
}

#endif

//EOF

