/**************************************************************************************

                                        Pager
                                    Implementation

**************************************************************************************/
//Tab Width: 4 Character:\t

#include "pager.h"	//Includes self, obviously

#include "cache.h"		//Includes cache and filesystem, because it uses those for its 
#include "filesystem.h"	//	functionality
#include "error.h"		//To allow logging of any and all sorts of itty bitty problems
						//	that we might or might not have

#include <string.h>		//String manipulation is very important

//-------------------------------------------------------------------------------------
//Main functions - defined in the header

uint8_t ndb_pagerInit(nDB* db) {
	return ndb_cacheInit(db);	//Not much to do here...
}

void ndb_pagerClose(nDB* db) {
	ndb_cacheClose(db);
}

//
//getPage!!
//

ndb_page* ndb_getPage_(nDB* db, NDB_PGNUM pg) {
	ndb_page* p;	
	
	assert(db);			//That would be a fail
	
	//Checks if the page is cached
	if ((p = ndb_cacheGet(db,pg))) {
		p->users++;	//Add a user to the page - so that the pointer never gets dereferenced
					//	while the page is in use
		return p;	//If it is in the cache, return it. That was easy.
	}
	
	//And now it gets a wee bit fun.
	//If the cache is maxed out (and non-0, since 0 is unlimited)
	//	the paging mechanism forces a clean - if it is still above the
	//	limit, return a cache_max error -> we do NOT want to flush the
	//	file automatically, because the user might want to leave all changes
	//	he/she does in memory
	if (db->cache_max && db->cache_max <= db->cache_amt) {
		ndb_cacheClean(db);
		//If the database is still full:
		if (db->cache_max <= db->cache_amt) {
			ndb_seterr(db,NDBERR_CACHE);
			return NULL;
		}
	}
	//If it gets here, the cache has room for a new page, and the page is not
	//	yet in the cache
	
	
	//TODO: This might be dangerous, since during cleaning another thread might
	//	go in to add the same page, making a memory leak...
	//FIXED: Added a later check if the page exists in the cache to make sure
	//	that it is not the case.
	
	if ((p = ndb_cacheNewPage(db))) {
		char tmp;		//A temporary variable to check error codes
		p->page = pg;	//Set the cache page's page number
		
		//If it has a file, load data from it. If not, then it is considered a blank
		//	page.
		if (db->file) {
			//Read the file
			db->r(db->pagestart+pg*db->pagesize,db->pagesize,p->data,db->file);
		}
		
		
		if (!(tmp = ndb_cacheAddPage(db,p))) {	//If adding the page was successful
			p->users++;	//The page is in use
			return p;
		}
		
		//Oh dear, the page was not added to the cache
		//This can be one of two reasons: One, it is multithreaded, and another thread
		//	beat this one to the punch 0 in which case all's cool, we delete the current
		//	page, and then just return the cached one. Or, an error has occurred, in which
		//	case an error code is already set, so we return the page and return NULL
		
		//Delete the current page
		ndb_cacheDelPage(db,p);
		
		if (tmp==1) {	//If the return value is 1, the page exists in the cache
			if ((p = ndb_cacheGet(db,pg))) {
				p->users++;
				return p;	//Return it
			}
			//In this case, what it is I don't even.
			ndb_seterr(db,NDBERR_UNKNOWN);
		}
		//The cacheAddPage already set the necessary error, so finish it up
	}
	return NULL;	
}

void ndb_freePage_(ndb_page* p) {
	assert(p);
	assert(p->users);	//There are users!
	//assert(!p->readers && !p->writing);
	//	This part CAN be true, since the page can be opened elsewhere
	p->users--;
	//if 0 users, read and write locks are nonexistent
}

void ndb_pgFlush(nDB* db, ndb_page* p) {
	if (ndb_pgIsModified(p)) {	//If the page is modified
		ndb_pgRead(p);			//Set a readlock on the page
		//Write the actual page
		db->w(db->pagestart+db->pagesize*p->page,db->pagesize,p->data,db->file);
		p->modified=0;			//The page is now not modified
		//TODO: add a function wrapping this, since we have one wrapping
		//	the check of the value!
		ndb_pgDone(p);			//Unset the readlock
	}
}

void ndb_flush_(nDB* db) {
	assert(db);
	if (db->file && !db->readonly) {	//The file must exist, and must be opened for write!
		ndb_page* p;	//Define a page
		ndb_cacheIter i=0;	//And a cache iterator
		
		//We iterate the cache to get the pages
		while ((p = ndb_cacheIterate(db,&i))) {
			assert(p);	//The cache is not supposed to return BS
			ndb_pgFlush(db,p);
		}	
	}
}

//-------------------------------------------------------------------------------------
//Pointer to page data
void* ndb_pgptr(ndb_page* p,uint16_t location) {
	assert(p);
	return (p->data+location);
}

//String access functions
char ndb_pgstrw_(ndb_page* p, uint16_t location, char* string) {
	assert(p && string);
	strcpy((char*)(p->data+location),string);
	return 0;
}
char* ndb_pgstrr_(ndb_page* p, uint16_t location) {
	char* str;
	assert(p);
	str = malloc(ndb_pgstrlen(p,location));	//Allocates a string
	strcpy(str,(char*)(p->data+location));
	return str;
}
uint16_t ndb_pgstrlen(ndb_page* p, uint16_t location) {
	assert(p);
	return (uint16_t)strlen((char*)(p->data+location));
}

//Object write/read functions
char ndb_pgw_(ndb_page* p, uint16_t location, uint16_t numbytes, void* object) {
	assert(p && object && numbytes);
	memcpy(p->data+location,object,numbytes);
	return 0;
}
char ndb_pgr_(ndb_page* p, uint16_t location, uint16_t numbytes, void* object) {
	assert(p && object && numbytes);
	memcpy(object,p->data+location,numbytes);
	return 0;
}

//Character access functions
uint8_t ndb_pggetc_(ndb_page* p, uint16_t location) {
	assert(p);					//The page has to exist
	return p->data[location];
}
char ndb_pgputc_(ndb_page* p, uint16_t location, uint8_t c) {
	assert(p);
	p->data[location]=c;
	return 0;
}

//Memsets, etc
char ndb_pgmemset_(ndb_page* p, uint16_t location, int value, uint16_t numbytes) {
	assert(p && numbytes);	//Numbytes shouldn't be 0...
	memset(p->data+location,value,numbytes);
	return 0;
}
char ndb_pgclear_(nDB* db, ndb_page* p) {
	assert(db && p);
	return ndb_pgmemset_(p,0,0,db->pagesize);
}

//Simplifier functions:
char ndb_pgwu16_(ndb_page* p, uint16_t location, uint16_t i) {
	return ndb_pgw_(p,location,sizeof(uint16_t),&i);
}
char ndb_pgwu32_(ndb_page* p, uint16_t location, uint32_t i) {
	return ndb_pgw_(p,location,sizeof(uint32_t),&i);
}
char ndb_pgwu64_(ndb_page* p, uint16_t location, uint64_t i) {
	return ndb_pgw_(p,location,sizeof(uint64_t),&i);
}

uint16_t ndb_pgru16_(ndb_page* p, uint16_t location) {
	uint16_t i;
	ndb_pgr_(p,location,sizeof(uint16_t),&i);
	return i;
}
uint32_t ndb_pgru32_(ndb_page* p, uint16_t location) {
	uint32_t i;
	ndb_pgr_(p,location,sizeof(uint32_t),&i);
	return i;
}
uint64_t ndb_pgru64_(ndb_page* p, uint16_t location) {
	uint64_t i;
	ndb_pgr_(p,location,sizeof(uint64_t),&i);
	return i;
}

//Label writer, for the lulz
void ndb_putlP_(ndb_page* p,uint8_t lbl) {
	assert(p);
	ndb_pgWriteLock_(p);
	ndb_pgputc_(p,0,lbl);
	ndb_pgWriteUnlock_(p);
}

void ndb_putl(nDB* db, NDB_PGNUM p, uint8_t lbl) {
	ndb_page* pg;
	
	assert(db);
	
	if ((pg=ndb_getPage(db,p))) {	//Only do stuff if no fail
		ndb_putlP(pg,lbl);
		ndb_freePage(pg);
	}
}

uint16_t ndb_pgCompare(ndb_page* p, uint16_t location, void* object, uint16_t size) {
	uint16_t i=0;	//Iterator
	assert(p && object && size);	//We need these to be non-NULL/0
	
	//Loops while they are equal
	while (i < size && ((uint8_t*)object)[i]==p->data[i+location]) ++i;
	
	if (i==size && ((uint8_t*)object)[i-1]==p->data[i-1+location]) return 0;
	else if (i) {
		if (((uint8_t*)object)[i-1] > p->data[i-1+location]) return 1;
		return -1;
	}
	if (((uint8_t*)object)[i] > p->data[i+location]) return 1;
	return -1;
	
	/*Fail code is fail
	
	if (((uint8_t*)object)[i-1] < p->data[i-1+location]) return -(i-1);
	else if (((uint8_t*)object)[i-1] > p->data[i-1+location]) return i-1;
	else return 0;	//They are equal
	*/
	
}

//-------------------------------------------------------------------------------------
//LOGGING

#if defined(NDB_PVER) && NDB_HASLOG
ndb_page* ndb_getPage__(nDB* db, uint64_t pg,char* loc,int line) {
	char c = (ndb_cacheGet(db,pg)?'C':'F');
	ndb_page* p = ndb_getPage_(db,pg);
	if (!p) {
		ndb_log(NDB_PVER,"(PGET   Pg:%-7lu| From:%c | Returned NULL | %s:%d )",pg,c,loc,line);
		return p;
	}
	ndb_log(NDB_PVER,"\n(PGET   Pg:%-7lu| U/MW/R:%2u/%i%i/%i | From:%c | %s:%d )",pg,
		p->users,p->modified,p->writing,p->readers,c,loc,line);
	return p;
}
void ndb_freePage__(ndb_page* p,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PVER,"\n(PFREE  Pg:%-7lu| U/MW/R:%2u/%i%i/%i | %s:%d )",p->page,
		p->users,p->modified,p->writing,p->readers,loc,line);
	ndb_freePage_(p);
}
void ndb_flush__(nDB* db,char* loc,int line) {
	ndb_log(NDB_PVER,"\n(DBFLUSH %s:%d )",loc,line);
	ndb_flush_(db);
}
#endif

#if defined(NDB_PGVER) && NDB_HASLOG
char ndb_pgstrw__(ndb_page* p, uint16_t location, char* string,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PSTRW  Pg:%-7lu| Loc:%-5u| %s:%d )",p->page,location,loc,line);
	return ndb_pgstrw_(p,location,string);
}
char* ndb_pgstrr__(ndb_page* p, uint16_t location,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PSTRR  Pg:%-7lu| Loc:%-5u| %s:%d )",p->page,location,loc,line);
	return ndb_pgstrr_(p,location);
}
char ndb_pgw__(ndb_page* p, uint16_t location, uint16_t numbytes, void* object,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PW    Pg:%-7lu| Loc:%-5u| Size:%-4u| %s:%d )",p->page,location,numbytes,loc,line);
	return ndb_pgw_(p,location,numbytes,object);
}
char ndb_pgr__(ndb_page* p, uint16_t location, uint16_t numbytes, void* object,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PR    Pg:%-7lu| Loc:%-4u| Size:%-4u| %s:%d )",p->page,location,numbytes,loc,line);
	return ndb_pgr_(p,location,numbytes,object);
}
uint8_t ndb_pggetc__(ndb_page* p, uint16_t location,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PGETC  Pg:%-7lu| Loc:%-5u| %s:%d )",p->page,location,loc,line);
	return ndb_pggetc_(p,location);
}
char ndb_pgputc__(ndb_page* p, uint16_t location, uint8_t c,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PPUTC  Pg:%-7lu| Loc:%-5u| Val:%c | %s:%d )",p->page,location,(c=='\0'?' ':c),loc,line);
	return ndb_pgputc_(p,location,c);
}
char ndb_pgmemset__(ndb_page* p, uint16_t location, int value, uint16_t numbytes,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PMEMSET Pg:%-7lu| Loc:%-5u| Num/Val:%u/%u | %s:%d )",p->page,location,numbytes,
		(unsigned int)value,loc,line);
	return ndb_pgmemset_(p,location,value,numbytes);
}
char ndb_pgclear__(nDB* db, ndb_page* p,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PCLEAR Pg:%-7lu| %s:%d )",p->page,loc,line);
	return ndb_pgclear_(db,p);
}
char ndb_pgwu16__(ndb_page* p, uint16_t location, uint16_t i,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PWU16  Pg:%-7lu| Loc:%-5u| Val:%u | %s:%d )",p->page,location,i,loc,line);
	return ndb_pgwu16_(p,location,i);
}
char ndb_pgwu32__(ndb_page* p, uint16_t location, uint32_t i,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PWU32  Pg:%-7lu| Loc:%-5u| Val:%u | %s:%d )",p->page,location,i,loc,line);
	return ndb_pgwu32_(p,location,i);
}
char ndb_pgwu64__(ndb_page* p, uint16_t location, uint64_t i,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PWU64  Pg:%-7lu| Loc:%-5u| Val:%lu | %s:%d )",p->page,location,i,loc,line);
	return ndb_pgwu64_(p,location,i);
}
uint16_t ndb_pgru16__(ndb_page* p, uint16_t location,char* loc,int line) {
	uint16_t i;
	assert(p);
	ndb_log(NDB_PGVER,"\n(PRU16  Pg:%-7lu| Loc:%-5u| ",p->page,location);
	i=ndb_pgru16_(p,location);
	ndb_log(NDB_PGVER,"Val:%u | %s:%d )",i,loc,line);
	return i;
}
uint32_t ndb_pgru32__(ndb_page* p, uint16_t location,char* loc,int line) {
	uint32_t i;
	assert(p);
	ndb_log(NDB_PGVER,"\n(PRU32  Pg:%-7lu| Loc:%-5u| ",p->page,location);
	i=ndb_pgru32_(p,location);
	ndb_log(NDB_PGVER,"Val:%u | %s:%d )",i,loc,line);
	return i;
}
uint64_t ndb_pgru64__(ndb_page* p, uint16_t location,char* loc,int line) {
	uint64_t i;
	assert(p);
	ndb_log(NDB_PGVER,"\n(PRU64  Pg:%-7lu| Loc:%-5u| ",p->page,location);
	i=ndb_pgru64_(p,location);
	ndb_log(NDB_PGVER,"Val:%lu | %s:%d )",i,loc,line);
	return i;
}
void ndb_putlP__(ndb_page* p,uint8_t lbl,char* loc,int line) {
	assert(p);
	ndb_log(NDB_PGVER,"\n(PUTL   Pg:%-7lu| Val:%c | %s:%i )",p->page,(lbl=='\0'?' ':lbl),loc,line);
	ndb_putlP_(p,lbl);
}

#endif

//EOF

